# Generated by Django 2.2.10 on 2020-02-27 21:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0019_eveasset'),
    ]

    operations = [
        migrations.RenameField(
            model_name='eveasset',
            old_name='item',
            new_name='item_name',
        ),
        migrations.AddField(
            model_name='eveasset',
            name='category_id',
            field=models.IntegerField(default=0),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='eveasset',
            name='group_id',
            field=models.IntegerField(default=0),
            preserve_default=False,
        ),
    ]



from django_eveonline_connector.models import EveClient, EveToken
from django_eveonline_connector.exceptions import EveDataResolutionError
from django.core.cache import cache
import logging

logger = logging.getLogger(__name__)

def resolve_id(id):
    return resolve_ids_with_types([95465499, id])[int(id)]

def resolve_ids(ids):
    resolved_ids = EveClient.call('post_universe_names', ids=ids).data 
    response = {}
    for resolved_id in resolved_ids:
        external_id = resolved_id['id']
        external_name = resolved_id['name']
        if external_id not in response:
            response[external_id] = external_name
    return response

def resolve_names(names):
    resolved_ids = EveClient.call('post_universe_ids', names=names).data
    return resolved_ids


def resolve_ids_with_types(ids):
    request = EveClient.call('post_universe_names', ids=ids)
    if request.status != 200:
        raise EveDataResolutionError(f"Failed to resolve IDs: {ids}")
    resolved_ids = request.data
    response = {}
    for resolved_id in resolved_ids:
        external_id = resolved_id['id']
        external_name = resolved_id['name']
        external_type = resolved_id['category']
        if external_id not in response:
            response[external_id] = {
                "name": external_name,
                "type": external_type
            }
    return response


def get_type_id(type_id):
    response = EveClient.call('get_universe_types_type_id', type_id=type_id, raise_exception=True)
    if response.status != 200:
        raise EveDataResolutionError(
            f"Failed to resolve type_id({type_id}) using ESI. Notify CCP.")
    return response.data

def get_station_id(station_id):
    return EveClient.call('get_universe_stations_station_id', station_id=station_id).data


def get_group_id(group_id):
    return EveClient.call('get_universe_groups_group_id', group_id=group_id).data

def get_category_id(category_id):
    return EveClient.call('get_universe_categories_category_id', category_id=category_id).data

def get_structure_id(structure_id, token_entity_id, safe=True):
    from django_eveonline_connector.models import EveStructure
    if str(structure_id) in cache:
        return cache.get(str(structure_id))

    if safe: # not running safe can lead to ESI lockouts, structure resolution is busted 
        structure = EveStructure.objects.filter(structure_id=structure_id).first()
        if structure:
            return structure.name
        else:
            return "Unknown Structure"
    else:
        try:
            token = EveToken.objects.get(evecharacter__external_id=token_entity_id)
        except EveToken.DoesNotExist:
            raise EveDataResolutionError('Attempted to resolve a structure without a token')

        response = EveClient.call('get_universe_structures_structure_id', token=token, structure_id=structure_id)
        if response.status != 200:
            raise EveDataResolutionError(f"Failed to resolve structure {structure_id} using external ID {token_entity_id}. Reason: {response.data}")

        if cache:
            cache.set(str(structure_id), response.data['name'])
            return response.data['name']
        else:
            return response.data['name']


def get_structure(structure_id, corporation_id):
    from django_eveonline_connector.models import EveCorporation
    corporation = EveCorporation.objects.get(external_id=corporation_id)
    response = EveClient.call(
        'get_universe_structures_structure_id', token=corporation.ceo.token, structure_id=structure_id)
    if response.status != 200:
        raise EveDataResolutionError(
            f"Failed to resolve structure {structure_id} using external ID {corporation_id}. Reason: {response.data}")

    return response



from django_eveonline_connector.tasks import update_tokens, pull_corporation_roster
from django_eveonline_connector.models import *
from django.contrib.auth.models import User, Group
from django.test import TestCase
from datetime import timedelta
from django.utils import timezone
import pytz
from unittest.mock import patch

class MockResponseObject():
    def __init__(self, status_code, data):
        self.status_code = status_code 
        self.data = data

def mock_update_corporations_for_pull_test():
    corporation = EveCorporation.objects.get_or_create(external_id=200, name="DUMMY CORPORATION")[0]
    character_2 = EveCharacter.objects.get(external_id=2)
    character_4 = EveCharacter.objects.get(external_id=4)
    character_2.corporation = corporation
    character_4.corporation = corporation
    character_2.save()
    character_4.save()
class UpdateTokenTest(TestCase):
    def setUp(self):
        self.client = EveClient.objects.create(esi_callback_url="TEST",
                                 esi_client_id="TEST",
                                 esi_secret_key="TEST")

    def tearDown(self):
        self.client.delete() 

    @patch('django_eveonline_connector.models.EveToken.refresh')
    def test_valid_token(self, mock_refresh):
        mock_refresh.return_value = True
        character = EveCharacter.objects.create(
            external_id=12345,
            name="Test Character"
        )
        token = EveToken.objects.create()
        character.token = token 
        character.save() 
        token.scopes.set(EveScope.objects.all())
        self.assertTrue(token.valid)
        update_tokens()
        self.assertTrue(EveToken.objects.all().count() == 1)
        self.assertTrue(EveToken.objects.all()[0].pk == token.pk)
        self.assertTrue(EveToken.objects.all()[0].invalidated == None)
        token.delete()

    @patch('django_eveonline_connector.models.EveToken.refresh')
    def test_invalidating_token(self, mock_refresh):
        mock_refresh.return_value = True
        character = EveCharacter.objects.create(
            external_id=12345,
            name="Test Character"
        )
        token = EveToken.objects.create()
        character.token = token 
        character.save()
        token.scopes.set(EveScope.objects.all())
        token.scopes.remove(EveScope.objects.filter(required=True)[0])
        self.assertFalse(token.valid)
        update_tokens()
        self.assertTrue(EveToken.objects.all().count() == 1)
        token = EveToken.objects.all()[0]
        self.assertTrue((timezone.now() - token.invalidated).days == 0)
        self.assertTrue(token.pk == token.pk)
        self.assertTrue(token.invalidated != None)

    @patch('django_eveonline_connector.models.EveToken.refresh')
    def test_delete_invalidated_token(self, mock_refresh):
        mock_refresh.return_value = True
        token = EveToken.objects.create()
        character = EveCharacter.objects.create(
            external_id=12345,
            name="Test Character"
        )
        character.token = token 
        character.save()
        token.scopes.set(EveScope.objects.all())
        token.scopes.remove(EveScope.objects.filter(required=True)[0])
        self.assertFalse(token.valid)
        update_tokens()
        token.invalidated = timezone.now() - timedelta(days=6)
        token.save()
        update_tokens()
        self.assertTrue(EveToken.objects.all().count() == 1)
        token.invalidated = timezone.now() - timedelta(days=8)
        token.save()
        update_tokens()
        self.assertTrue(EveToken.objects.all().count() == 0)

class PullCorporationRoster(TestCase):
    def setUp(self):
        self.corporation = EveCorporation.objects.create(external_id=100, name="TEST CORPORATION")
        EveCharacter.objects.create(
            external_id=1, corporation=self.corporation)
        EveCharacter.objects.create(
            external_id=2, corporation=self.corporation)
        EveCharacter.objects.create(
            external_id=3, corporation=self.corporation)
        EveCharacter.objects.create(
            external_id=4, corporation=self.corporation)
        EveCharacter.objects.create(
            external_id=5, corporation=self.corporation)

    @patch('django_eveonline_connector.tasks.EveClient.call')
    @patch('django_eveonline_connector.models.EveCharacter.update_character_corporation')
    def test_pull_roster(self, mock_character_update, mock_eve_client):
        mock_character_update.return_value = MockResponseObject(status_code=200, data=[200])
        mock_character_update.side_effect = mock_update_corporations_for_pull_test()
        mock_eve_client.return_value = MockResponseObject(status_code=200, data=[1,3,5])
        pull_corporation_roster(self.corporation.external_id)
        characters = EveCharacter.objects.filter(corporation=self.corporation)
        character_ids = characters.values_list('external_id', flat=True)
        self.assertTrue(list(character_ids) == [1,3,5])


class AssignEveGroupsTest(TestCase):
    def setUp(self):
        self.group = Group.objects.create(name="TEST")
        self.eve_group_rule = EveGroupRule.objects.create(group=self.group)
        self.user = User.objects.create(username="TEST USER")
        self.excluded_user = User.objects.create(username="EXCLUDED USER")
        self.excluded_user_2 = User.objects.create(username="EXCLUDED USER 2")
        self.eve_token = EveToken.objects.create(user=self.user)
        self.eve_alliance = EveAlliance.objects.create(
            name="TEST ALLIANCE PLEASE IGNORE",
            external_id=912345
        )
        self.eve_corporation = EveCorporation.objects.create(
            name="TEST CORPORATION",
            alliance=self.eve_alliance,
            external_id=512345
        )
        self.eve_character = EveCharacter.objects.create(name="TEST",
            token=self.eve_token,
            corporation=self.eve_corporation,
            external_id=112345)

    def tearDown(self):
        self.group.delete()
        self.eve_group_rule.delete()
        self.eve_token.delete()
        self.eve_alliance.delete()
        self.eve_corporation.delete()
        self.eve_character.delete()

    def test_group_assignment(self):
        from django_eveonline_connector.tasks import assign_eve_groups
        self.excluded_user.groups.add(self.group)
        self.excluded_user_2.groups.add(self.group)
        self.eve_group_rule.characters.add(self.eve_character)
        assign_eve_groups()
        self.assertTrue(self.group not in self.excluded_user.groups.all())
        self.assertTrue(self.group not in self.excluded_user_2.groups.all())
        self.assertTrue(self.group in self.user.groups.all())



# Generated by Django 2.2.13 on 2020-09-20 21:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0026_auto_20200309_1419'),
    ]

    operations = [
        migrations.AddField(
            model_name='evescope',
            name='required',
            field=models.BooleanField(default=True),
        ),
        migrations.DeleteModel(
            name='EveTokenType',
        ),
    ]



from django.shortcuts import render, redirect
from django_eveonline_connector.models import EveCorporation, EveCharacter, PrimaryEveCharacterAssociation, EveStructure
from django.contrib import messages
from django.db.models import Q 
from itertools import chain
from django.contrib.auth.decorators import login_required, permission_required

import logging
logger = logging.getLogger(__name__)

@login_required
@permission_required('django_eveonline_connector.view_evecorporation', raise_exception=True)
def view_corporations(request):
    return render(request, 'django_eveonline_connector/adminlte/corporations/list_corporations.html', context={
        'corporations': EveCorporation.objects.filter(track_corporation=True)
    })


@login_required
@permission_required('django_eveonline_connector.view_evecorporation', raise_exception=True)
def view_corporation(request, external_id):
    context = {}
    context['corporation'] = EveCorporation.objects.get(external_id=external_id)
    context['characters'] = PrimaryEveCharacterAssociation.objects.filter(
        character__corporation=context['corporation'])
    context['unauthorized_characters'] = list(
        chain(
            EveCharacter.objects.filter(corporation=context['corporation'], token=None),
            EveCharacter.objects.filter(corporation=context['corporation']).exclude(Q(token__invalidated=None))
        )
    )
    return render(request, 'django_eveonline_connector/adminlte/corporations/view_corporation_roster.html', context)


@login_required
@permission_required('django_eveonline_connector.view_evecorporation', raise_exception=True)
def view_corporation_structures(request, external_id):
    if request.user.primary_evecharacter.character.corporation.external_id == external_id and not request.user.has_perm('django_eveonline_connector.bypass_corporation_view_requirements'):
        messages.error(request, "You do not have permission to view those structures, as you are not a member of that corporation or are missing the BYPASS permission")
        return redirect('django-eveonline-connector-view-corporation', external_id)
    context = {
        'corporation': EveCorporation.objects.get(external_id=external_id),
        'structures': EveStructure.objects.filter(entity__external_id=external_id)
    }
    return render(request, 'django_eveonline_connector/adminlte/corporations/view_corporation_structures.html', context)




# Generated by Django 2.2.10 on 2020-03-03 18:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0021_auto_20200227_2343'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='EveClone',
            new_name='EveJumpClone',
        ),
    ]



from django.urls import resolve
from django.shortcuts import redirect
from django.urls import reverse
from django.contrib import messages
from django_eveonline_connector.models import EveScope

def required_scopes(scopes=[]):
    def _method_wrapper(view_method):
        def wrap(request, *args, **kwargs):
            if not request.user.primary_evecharacter:
                messages.error(request, "You are missing an EVE Online character.")
                return redirect('/')
            redirect_user = False 
            for scope in scopes: 
                scope = EveScope.objects.get_or_create(name=scope)[0]
                if scope not in request.user.primary_evecharacter.character.token.scopes.all():
                    redirect_user = True 
                    request.user.primary_evecharacter.character.token.requested_scopes.add(scope)
            if redirect_user:
                messages.warning(request, "You are missing required scopes. Check your token and try again.")
                return redirect("/")
            return view_method(request, *args, **kwargs)
        return wrap 
    return _method_wrapper



# Generated by Django 2.2.13 on 2020-10-26 19:20

from django.db import migrations, models
import django.db.models.deletion


def load_required_roles(apps, schema_editor):
    EveCorporationRole = apps.get_model('django_eveonline_connector',
                                        'EveCorporationRole')
    corporation_roles = (
        ("Account_Take_1", "Account Take 1"),
        ("Account_Take_2", "Account Take 2"),
        ("Account_Take_3", "Account Take 3"),
        ("Account_Take_4", "Account Take 4"),
        ("Account_Take_5", "Account Take 5"),
        ("Account_Take_6", "Account Take 6"),
        ("Account_Take_7", "Account Take 7"),
        ("Accountant", "Accountant"),
        ("Auditor", "Auditor"),
        ("Communications_Officer", "Communications Officer"),
        ("Config_Equipment", "Config Equipment"),
        ("Config_Starbase_Equipment", "Config Starbase Equipment"),
        ("Container_Take_1", "Container Take 1"),
        ("Container_Take_2", "Container Take 2"),
        ("Container_Take_3", "Container Take 3"),
        ("Container_Take_4", "Container Take 4"),
        ("Container_Take_5", "Container Take 5"),
        ("Container_Take_6", "Container Take 6"),
        ("Container_Take_7", "Container Take 7"),
        ("Contract_Manager", "Contract Manager"),
        ("Diplomat", "Diplomat"),
        ("Director", "Director"),
        ("Factory_Manager", "Factory Manager"),
        ("Fitting_Manager", "Fitting Manager"),
        ("Hangar_Query_1", "Hangar Query 1"),
        ("Hangar_Query_2", "Hangar Query 2"),
        ("Hangar_Query_3", "Hangar Query 3"),
        ("Hangar_Query_4", "Hangar Query 4"),
        ("Hangar_Query_5", "Hangar Query 5"),
        ("Hangar_Query_6", "Hangar Query 6"),
        ("Hangar_Query_7", "Hangar Query 7"),
        ("Hangar_Take_1", "Hangar Take 1"),
        ("Hangar_Take_2", "Hangar Take 2"),
        ("Hangar_Take_3", "Hangar Take 3"),
        ("Hangar_Take_4", "Hangar Take 4"),
        ("Hangar_Take_5", "Hangar Take 5"),
        ("Hangar_Take_6", "Hangar Take 6"),
        ("Hangar_Take_7", "Hangar Take 7"),
        ("Junior_Accountant", "Junior Accountant"),
        ("Personnel_Manager", "Personnel Manager"),
        ("Rent_Factory_Facility", "Rent Factory Facility"),
        ("Rent_Office", "Rent Office"),
        ("Rent_Research_Facility", "Rent Research Facility"),
        ("Security_Officer", "Security Officer"),
        ("Starbase_Defense_Operator", "Starbase Defense Operator"),
        ("Starbase_Fuel_Technician", "Starbase Fuel Technician"),
        ("Station_Manager", "Station Manager"),
        ("Trader", "Trader"),
    )
    for role_tuple in corporation_roles:
        EveCorporationRole.objects.create(
            codename=role_tuple[0],
            name=role_tuple[1],
        )


def unload_required_roles(apps, schema_editor):
    EveCorporationRole = apps.get_model(
        'django_eveonline_connector', 'EveCorporationRole')
    EveCorporationRole.objects.all().delete()  # dangerous but it'll be fine

class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
        ('django_eveonline_connector', '0038_auto_20201025_1808'),
    ]

    operations = [
        migrations.CreateModel(
            name='EveCorporationRole',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('codename', models.CharField(max_length=64, unique=True)),
                ('name', models.CharField(max_length=64, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='EveGroupRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('include_alts', models.BooleanField(default=True)),
                ('alliances', models.ManyToManyField(blank=True, to='django_eveonline_connector.EveAlliance')),
                ('characters', models.ManyToManyField(blank=True, to='django_eveonline_connector.EveCharacter')),
                ('corporations', models.ManyToManyField(blank=True, to='django_eveonline_connector.EveCorporation')),
                ('group', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='auth.Group')),
                ('roles', models.ManyToManyField(blank=True, to='django_eveonline_connector.EveCorporationRole')),
            ],
        ),
        migrations.AddField(
            model_name='evecharacter',
            name='roles',
            field=models.ManyToManyField(blank=True, to='django_eveonline_connector.EveCorporationRole'),
        ),
        migrations.RunPython(
            load_required_roles, reverse_code=unload_required_roles
        ),
    ]



# Generated by Django 2.2.4 on 2019-12-10 14:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0007_auto_20191210_1430'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eveclient',
            name='esi_sso_url',
            field=models.URLField(editable=False, max_length=512),
        ),
    ]



# Generated by Django 2.2.4 on 2019-09-03 20:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='evetoken',
            name='primary',
            field=models.BooleanField(default=False),
        ),
    ]



class EveDataResolutionError(Exception):
    """
    Thrown when we failed to resolve something using the static database or ESI
    This should not happen, it usually means that we have a fatal problem in our code OR CCP has an issue with ESI or the static export. 
    """
    def __init__(self, msg):
        self.msg = msg

class EveMissingScopeException(Exception):
    """
    Thrown when we make a call without a proper token. 
    """
    def __init__(self, msg):
        self.msg = msg 


# Generated by Django 2.2.10 on 2020-02-27 18:21

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0017_primaryevecharacterassociation'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='evetoken',
            name='primary',
        ),
    ]



# Generated by Django 2.2.10 on 2020-03-09 14:19

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0025_auto_20200308_2220'),
    ]

    operations = [
        migrations.RenameField(
            model_name='evetransaction',
            old_name='type_name',
            new_name='item_name',
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-27 18:14

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('django_eveonline_connector', '0016_auto_20200226_1730'),
    ]

    operations = [
        migrations.CreateModel(
            name='PrimaryEveCharacterAssociation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('character', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='primary_to', to='django_eveonline_connector.EveCharacter')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='primary_evecharacter', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from celery import task, shared_task
from .models import *
from django.utils import timezone
from django.db.models import Q

import logging
import pytz
logger = logging.getLogger(__name__)

"""
Global Tasks
These are what users register to maintain up-to-date EveEntity information.
"""
@shared_task
def update_affiliations():
    """
    Update the affiliations (character corporations, corporation alliances)
    Cached for 3600 seconds, so don't run it more frequently than that. 
    """
    character_ids = EveCharacter.objects.all().values_list('external_id', flat=True)
    affiliations = EveClient.call(
        'post_characters_affiliation', characters=character_ids).data

    for affiliation in affiliations:
        try:
            character = EveCharacter.objects.get(
                external_id=affiliation['character_id'])
            if not EveCorporation.objects.filter(external_id=affiliation['corporation_id']):
                EveCorporation.create_from_external_id(
                    affiliation['corporation_id'])
            corporation = EveCorporation.objects.get(
                external_id=affiliation['corporation_id'])
            if 'alliance_id' in affiliation:
                if not EveAlliance.objects.filter(external_id=affiliation['alliance_id']):
                    EveAlliance.create_from_external_id(affiliation['alliance_id'])
                alliance = EveAlliance.objects.get(external_id=affiliation['alliance_id'])
            else:
                alliance = None 

            character.corporation = corporation
            character.save()
            if alliance:
                corporation.alliance = alliance
                corporation.save()
        except Exception as e:
            logger.error("Failed to update affiliation: %s" % affiliation)
            logger.exception(e)

@shared_task
def update_tokens():
    EveToken.objects.filter(evecharacter=None).delete()
    for token in EveToken.objects.all():
        token.refresh()

    for token in EveToken.objects.all():
        if token.valid and token.invalidated:
            token.invalidated = None 
            token.save() 
        elif not token.valid and not token.invalidated:
            token.invalidated = timezone.now()
            token.save()
        elif token.invalidated:
            time_passed = timezone.now() - token.invalidated
            if time_passed.days > 7:
                token.delete()

@shared_task
def update_characters(jitter_max=1800):
    for eve_character in EveCharacter.objects.all():
        if eve_character.token and eve_character.token.valid and eve_character.corporation.track_characters: 
            logger.info(f"Queueing batch update tasks for {eve_character.name}")
            jitter = (eve_character.pk*10) % jitter_max
            update_character_assets.apply_async(
                args=[eve_character.external_id],
                countdown=jitter)
                
            update_character_contacts.apply_async(
                args=[eve_character.external_id],
                countdown=jitter)
                
            update_character_contracts.apply_async(
                args=[eve_character.external_id],
                countdown=jitter)
                
            update_character_journal.apply_async(
                args=[eve_character.external_id],
                countdown=jitter)
                
            update_character_jumpclones.apply_async(
                args=[eve_character.external_id],
                countdown=jitter)
                
            update_character_skills.apply_async(
                args=[eve_character.external_id],
                countdown=jitter)
                
            update_character_transactions.apply_async(
                args=[eve_character.external_id],
                countdown=jitter)

@shared_task
def update_character(character_id):
    eve_character = EveCharacter.objects.get(character_id)
    eve_character.update_character_corporation()

    update_character_assets.apply_async(
        args=[eve_character.external_id],
        countdown=jitter)

    update_character_contacts.apply_async(
        args=[eve_character.external_id],
        countdown=jitter)

    update_character_contracts.apply_async(
        args=[eve_character.external_id],
        countdown=jitter)

    update_character_journal.apply_async(
        args=[eve_character.external_id],
        countdown=jitter)

    update_character_jumpclones.apply_async(
        args=[eve_character.external_id],
        countdown=jitter)

    update_character_skills.apply_async(
        args=[eve_character.external_id],
        countdown=jitter)

    update_character_transactions.apply_async(
        args=[eve_character.external_id],
        countdown=jitter)

@shared_task
def update_character_roles():
    for eve_character in EveCharacter.objects.all():
        if eve_character.token and eve_character.token.valid:
            update_character_corporation_roles.apply_async(args=[eve_character.external_id])

@shared_task
def assign_eve_groups():
    for group_rule in EveGroupRule.objects.all():
        for user in group_rule.invalid_user_list:
            logger.info(
                f"Removing group ({group_rule.group} to user ({user})")
            user.groups.remove(group_rule.group)
        for user in group_rule.missing_user_list:
            logger.info(
                f"Adding group ({group_rule.group} to user ({user})")
            user.groups.add(group_rule.group)

@shared_task
def update_corporations():
    for eve_corporation in EveCorporation.objects.all():
        if eve_corporation.evecharacter_set.all().count() == 0:
            eve_corporation.delete()
            continue
        if eve_corporation.track_corporation:
            pull_corporation_roster.apply_async(args=[eve_corporation.external_id])
            update_corporation_alliance.apply_async(
                args=[eve_corporation.external_id])
            update_corporation_ceo.apply_async(
                args=[eve_corporation.external_id])
        else:
            logger.info(f"Skipping corporation update for {eve_corporation.name}: Not Tracked")

@shared_task
def update_alliances():
    for eve_alliance in EveAlliance.objects.all():
        update_alliance_executor.apply_async(args=[eve_alliance.external_id])

@shared_task
def update_structures():
    for eve_corporation in EveCorporation.objects.filter(track_corporation=True):
        try:
            update_corporation_eveentitydata(
                'corporations_corporation_id_structures', EveStructure, eve_corporation.external_id)
        except Exception as e:
            logger.error(e)
"""
Character tasks 
"""
def update_character_eveentitydata(op, data_model, character_id, delete=False):
    """
    Helper method for update_character_??? tasks.
    They basically all follow the same behavior. 
    """
    character = EveCharacter.objects.get(external_id=character_id)

    response = EveClient.call(op, character_id=character.external_id)

    if response.status != 200:
        logger.error(f"Failed to batch update {data_model} for {character_id}: {response.header} {response.data}")
        return

    items = response.data

    if 'X-Pages' in response.header and response.header['X-Pages'][0] > 1:
        for num in range(2, response.header['X-Pages'][0]+1):
            response = EveClient.call(
                op, character_id=character.external_id, page=num)
            if response.status != 200:
                continue 
            items += response.data 

    if len(items) == 0:
        return []

    if delete:
        data_model.objects.filter(entity=character).delete()

    data_model.create_from_esi_response(items, character.external_id)
    
    return items 

@shared_task
def update_character_assets(character_id, *args, **kwargs):
    op = 'get_characters_character_id_assets'
    data_model = EveAsset 
    update_character_eveentitydata(
        op, *args, **kwargs, character_id=character_id, delete=True, data_model=data_model)
    

@shared_task
def update_character_jumpclones(character_id, *args, **kwargs):
    op = 'get_characters_character_id_clones'
    data_model = EveJumpClone
    update_character_eveentitydata(op, *args, **kwargs, character_id=character_id, data_model=data_model, delete=True)


@shared_task 
def update_character_contacts(character_id, *args, **kwargs):
    op = 'get_characters_character_id_contacts'
    data_model = EveContact
    update_character_eveentitydata(
        op, *args, **kwargs, character_id=character_id, data_model=data_model, delete=True)

@shared_task
def update_character_contracts(character_id, *args, **kwargs):
    op = 'get_characters_character_id_contracts'
    data_model = EveContract
    update_character_eveentitydata(
        op, *args, delete=False, **kwargs, character_id=character_id, data_model=data_model)


@shared_task
def update_character_skills(character_id, *args, **kwargs):
    op = 'get_characters_character_id_skills'
    data_model = EveSkill
    response = update_character_eveentitydata(
        op, *args, **kwargs, character_id=character_id, data_model=data_model, delete=True)

    character = EveCharacter.objects.get(external_id=character_id)
    info = EveCharacterInfo.objects.get_or_create(character=character)[0]
    info.skill_points = response['total_sp']
    info.save()
    

@shared_task
def update_character_journal(character_id, *args, **kwargs):
    op = 'get_characters_character_id_wallet_journal'
    data_model = EveJournalEntry
    update_character_eveentitydata(
        op, *args, delete=False, **kwargs, character_id=character_id, data_model=data_model)

@shared_task
def update_character_transactions(character_id, *args, **kwargs):
    op = 'get_characters_character_id_wallet_transactions'
    data_model = EveTransaction
    update_character_eveentitydata(
        op, *args, delete=False, **kwargs, character_id=character_id, data_model=data_model)

@shared_task
def update_character_corporation_roles(character_id):
    character = EveCharacter.objects.get(external_id=character_id)
    eve_client = EveClient.get_instance()
    response = eve_client.call(op='get_characters_character_id_roles', character_id=character_id)
    if response.status != 200:
        logger.error(f"Failed to pull corporation roles for character {character_id}")
        return 
    roles = EveCorporationRole.objects.filter(
        codename__in=response.data['roles'])
    if roles != character.roles.all():
        character.roles.set(roles)
"""
Corporatoin Tasks 
"""


def update_corporation_eveentitydata(op, data_model, corporation_id, delete=False):
    """
    Helper method for update_corporation_??? tasks.
    They basically all follow the same behavior. 
    """

    corporation = EveCorporation.objects.get(external_id=corporation_id)

    response = EveClient.call(op, corporation_id=corporation.external_id)

    if response.status != 200:
        logger.error(
            f"Failed to batch update {data_model} for {character_id}: {response.header} {response.data}")
        return


    items = response.data

    if 'X-Pages' in response.header and response.header['X-Pages'][0] > 1:
        for num in range(2, response.header['X-Pages'][0]+1):
            response = EveClient.call(
                op, corporation_id=corporation.external_id, page=num)
            if response.status != 200:
                continue
            items += response.data

    if len(items) == 0:
        return []

    if delete:
        data_model.objects.filter(entity=corporation).delete()

    data_model.create_from_esi_response(items, corporation.external_id)

    return items

"""
Helper Tasks
"""
@shared_task
def update_corporation_alliance(corporation_id):
    """
    Legacy method that will be deprecated. Use update_affiliations().
    """
    esi_operation = EveClient.get_esi_app(
    ).op['get_corporations_corporation_id'](corporation_id=corporation_id)
    response = EveClient.get_esi_client().request(esi_operation)

    if 'alliance_id' not in response.data:
        logger.info(
            "No alliance found for corporation %s, returning None." % corporation_id)
        return None

    alliance_id = response.data['alliance_id']
    eve_corporation = EveCorporation.objects.get(external_id=corporation_id)

    if EveAlliance.objects.filter(external_id=alliance_id).exists():
        eve_corporation.alliance = EveAlliance.objects.get(
            external_id=alliance_id)
    else:
        eve_corporation.alliance = EveAlliance.create_from_external_id(
            alliance_id)

    eve_corporation.save()

@shared_task
def update_corporation_ceo(corporation_id):
    esi_operation = EveClient.get_esi_app(
    ).op['get_corporations_corporation_id'](corporation_id=corporation_id)
    response = EveClient.get_esi_client().request(esi_operation)

    ceo_id = response.data['ceo_id']
    eve_corporation = EveCorporation.objects.get(external_id=corporation_id)
    if EveCharacter.objects.filter(external_id=ceo_id).exists():
        eve_corporation.ceo = EveCharacter.objects.get(external_id=ceo_id)
    else:
       logger.info(f"Skipping CEO update for {eve_corporation.external_id}: CEO not in database")

    eve_corporation.save()

@shared_task
def pull_corporation_roster(corporation_id):
    token = EveToken.objects.filter(evecharacter__corporation__external_id=corporation_id).first()
    roster = EveClient.call('get_corporations_corporation_id_members', token=token, corporation_id=corporation_id)
    members_to_update = EveCharacter.objects.filter(~Q(external_id__in=roster.data))
    ids_that_exist = EveCharacter.objects.filter(external_id__in=roster.data).values_list('external_id', flat=True)

    for character in roster.data:
        if character not in ids_that_exist:
            EveCharacter.create_from_external_id(character)
        else:
            logger.info(f"skipping {character} due to already existing")

    for member in members_to_update:
        member.update_character_corporation()

@shared_task
def update_alliance_executor(alliance_id):
    esi_operation = EveClient.get_esi_app(
    ).op['get_alliances_alliance_id'](alliance_id=alliance_id)
    response = EveClient.get_esi_client().request(esi_operation)

    if 'executor_corporation_id' not in response.data:
        logger.info(
            "No executor found for alliance %s (likely closed). Returning None." % alliance_id)
        return None

    executor_id = response.data['executor_corporation_id']
    eve_alliance = EveAlliance.objects.get(external_id=alliance_id)

    if EveCorporation.objects.filter(external_id=executor_id).exists():
        eve_alliance.executor = EveCorporation.objects.get(
            external_id=executor_id)
    else:
        eve_alliance.executor = EveCorporation.create_from_external_id(
            executor_id)

    eve_alliance.save()



from esipy import EsiClient, EsiSecurity, EsiApp
from django.core.cache import cache
from django.db import models
from django.utils import timezone
from django.contrib.auth.models import User, Group 
from django.utils.dateparse import parse_datetime
from django.apps import apps
from django_singleton_admin.models import DjangoSingleton
from django_eveonline_connector.exceptions import EveMissingScopeException
import datetime
import logging
import json
import traceback
import pyswagger
from django.db.models import Q
logger = logging.getLogger(__name__)
app_config = apps.get_app_config('django_eveonline_connector')

id_types = (
    ("character", "character"),
    ("corporation", "corporation"),
    ("alliance", "alliance"),
)

"""
ESI Models
These models are used as wrappers around EsiPy
"""
class EveClient(DjangoSingleton):
    esi_base_url = models.URLField(
        default="https://esi.evetech.net/latest/swagger.json?datasource=tranquility")
    esi_callback_url = models.URLField()
    esi_client_id = models.CharField(max_length=255)
    esi_secret_key = models.CharField(max_length=255)

    def save(self, *args, **kwargs):
        
        super(EveClient, self).save(*args, **kwargs)

    def __str__(self):
        return str({
            'esi_base_url': self.esi_base_url,
            'esi_callback_url': self.esi_callback_url,
            'esi_client_id': self.esi_client_id
        })

    def as_dict(self):
        return {
            'esi_base_url': self.esi_base_url,
            'esi_callback_url': self.esi_callback_url,
            'esi_client_id': self.esi_client_id
        }

    def get_sso_url(self, extra_scopes=[]):
        """
        Expects a list of scope names 
        e.g ['esi.scope.v1', 'esi.scope.v2']
        """
        scope_list = EveScope.convert_to_list(EveScope.objects.filter(
            Q(required=True) | 
            Q(name__in=extra_scopes)
        ))
        scope_string = ",".join(scope_list)
        scope_string = f"esi_sso_url:{scope_string}"
        if scope_string in cache:
            return cache.get(scope_string)
        else:
            esi_sso_url = EsiSecurity(
                client_id=self.esi_client_id,
                redirect_uri=self.esi_callback_url,
                secret_key=self.esi_secret_key,
                headers={'User-Agent': "Krypted Platform"}
                ).get_auth_uri(scopes=scope_list,
                        state=self.esi_client_id)
            cache.set(scope_string, esi_sso_url, timeout=86400)
            return esi_sso_url

    @staticmethod
    def call(op, raise_exception=False, **kwargs):
        operation = EveClient.get_esi_app().op[op]
        # Pull the required scopes from the ESI Swagger Definition
        required_scopes = set()
        if operation.security:
            for definition in operation.security:
                if 'evesso' in definition:
                    for scope_name in definition['evesso']:
                        scope_creation = EveScope.objects.get_or_create(
                            name=scope_name)
                        scope = scope_creation[0]
                        if scope_creation[1]:
                            logger.error(
                                "Encountered unknown scope '%s', please notify Krypted developers." % scope)
                        required_scopes.add(scope)

        # Find the token with proper scopes and access
        if required_scopes:
            if 'token' in kwargs:
                token = kwargs.get('token')
                for scope in required_scopes:
                    if scope not in token.scopes.all():
                        raise EveMissingScopeException(f"EveClient was passed a token that is missing the required scopes: {scope}")

            elif 'character_id' in kwargs:
                try:
                    token = EveToken.objects.get(evecharacter__external_id=kwargs.get(
                        'character_id'), scopes__in=required_scopes)
                except EveToken.DoesNotExist as e:
                    raise EveMissingScopeException(
                        f"EveClient was passed a character_id that does not have the proper token: {required_scopes}")
            elif 'corporation_id' in kwargs:
                eve_corporation = EveCorporation.objects.get(
                    external_id=kwargs.get('corporation_id'))

                if not eve_corporation.ceo:
                    raise EveMissingScopeException(
                        'EveClient was passed a corporation_id that does not have a CEO')

                try:
                    token_pk = eve_corporation.ceo.token.pk
                    token = EveToken.objects.get(
                        pk=token_pk, scopes__in=required_scopes)
                except EveToken.DoesNotExist as e:
                    raise EveMissingScopeException(
                        'EveClient was passed a corporation_id that does not have the proper CEO token')
            
            else:
                raise Exception(
                    "Attempted to make protected EsiCall without valid token or auto-matching keyword argument")
        else:
            token = None

        if token:
            token.refresh()
            logger.info(f"Calling ESI: {op} with arguments {kwargs}")
            request = EveClient.get_esi_client(token=token).request(
                operation(**kwargs), raise_on_error=raise_exception)
        else:
            logger.info(f"Calling ESI: {op} with arguments {kwargs}")
            request = EveClient.get_esi_client().request(operation(**kwargs), raise_on_error=raise_exception)
        
        if request.status not in [200, 204]:
            logger.warning(f"Failed ({request.status}) ESI call '{op}' with {kwargs}. Response: {request.data}")

        return request 
    @staticmethod
    def get_required_scopes(op):
        operation = EveClient.get_esi_app().op[op]
        # Pull the required scopes from the ESI Swagger Definition
        required_scopes = set()
        if operation.security:
            for definition in operation.security:
                if 'evesso' in definition:
                    for scope_name in definition['evesso']:
                        scope_creation = EveScope.objects.get_or_create(
                            name=scope_name)
                        scope = scope_creation[0]
                        if scope_creation[1]:
                            logger.error(
                                "Encountered unknown scope '%s', please notify Krypted developers." % scope)
                        required_scopes.add(scope)
        return required_scopes
        
    @staticmethod
    def get_instance():
        app_config = apps.get_app_config('django_eveonline_connector')
        if app_config.ESI_SECRET_KEY and app_config.ESI_CLIENT_ID and app_config.ESI_CALLBACK_URL and app_config.ESI_BASE_URL:
            return EveClient(
                esi_client_id=app_config.ESI_CLIENT_ID,
                esi_secret_key=app_config.ESI_SECRET_KEY,
                esi_callback_url=app_config.ESI_CALLBACK_URL,
                esi_base_url=app_config.ESI_BASE_URL)

        elif EveClient.objects.all():
            return EveClient.objects.all()[0]

        raise Exception("EveClient is not configured.")

    @staticmethod
    def get_esi_app():
        """
        EsiApp is used to get operations for Eve Swagger Interface 
        """
        if 'esi_app' in cache:
            return cache.get('esi_app')
        else:
            esi_app = EsiApp(cache_time=86400).get_latest_swagger
            cache.set('esi_app', esi_app, timeout=86400)
            return esi_app

    @staticmethod
    def get_esi_client(token=None):
        """
        EsiClient is used to call operations for Eve Swagger Interface 
        """
        if token:
            return EsiClient(security=EveClient.get_esi_security(token), headers={'User-Agent': "Krypted Platform"})
        else:
            return EsiClient(headers={'User-Agent': "Krypted Platform"})

    @staticmethod
    def get_esi_security(token=None):
        """
        EsiSecurity is used to refresh and manage EVE Token objects
        """
        client = EveClient.get_instance()
        esi_security = EsiSecurity(
            redirect_uri=client.esi_callback_url,
            client_id=client.esi_client_id,
            secret_key=client.esi_secret_key,
            headers={'User-Agent': "Krypted Platform"})
        if token:
            esi_security.update_token(token.populate())
        return esi_security

    class Meta:
        verbose_name = "Eve Settings"
        verbose_name_plural = "Eve Settings"

"""
EVE SSO Models 
These models are used for the EVE Online token system
"""

# TODO: remove type, all scopes added.. add scopes via fixtures 

class EveScope(models.Model):
    name = models.CharField(unique=True, max_length=128)
    required = models.BooleanField(default=False)

    def __str__(self):
        return self.name

    @staticmethod
    def convert_to_list(scopes):
        return [scope.name for scope in scopes]

    @staticmethod
    def convert_to_string(scopes):
        return ",".join([scope.name for scope in scopes])

    @staticmethod
    def get_required():
        return EveScope.objects.filter(required=True)

class EveToken(models.Model):
    access_token = models.TextField()
    refresh_token = models.TextField()
    expires_in = models.IntegerField(default=0)
    expiry = models.DateTimeField(auto_now_add=True)
    invalidated = models.DateTimeField(blank=True, null=True, default=None)
    scopes = models.ManyToManyField("EveScope", blank=True)
    requested_scopes = models.ManyToManyField("EveScope", blank=True, related_name="requested_scopes", default=EveScope.get_required)
    user = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name="eve_tokens", null=True)

    def __str__(self):
        try:
            return "<%s:%s>" % (self.evecharacter.name, self.user)
        except Exception as e:
            return "<%s:%s>" % ("Unknown Character", self.user)

    @property
    def valid(self):
        for scope in self.requested_scopes.all():
            if scope not in self.scopes.all():
                return False 

        for scope in EveScope.objects.filter(required=True):
            if scope not in self.scopes.all():
                return False 

        if self.invalidated:
            return False
        return True 

    def refresh(self):
        esi_security = EveClient.get_esi_security()
        esi_security.update_token(self.populate())

        try:
            new_token = esi_security.refresh()
        except Exception as e:
            if b"invalid_grant" in e.response:
                if not self.invalidated:
                    self.invalidated = timezone.now()
                    self.save()
                return False    

        if timezone.now() > self.expiry:
            if self.invalidated:
                self.invalidated = None
            self.access_token = new_token['access_token']
            self.refresh_token = new_token['refresh_token']
            self.expiry = timezone.now() + datetime.timedelta(0,
                                                              new_token['expires_in'])
            self.save()
            return True
        else:
            logger.info("Token refresh not needed")
            return True

    def populate(self):
        data = {}
        data['access_token'] = self.access_token
        data['refresh_token'] = self.refresh_token
        data['expires_in'] = self.expires_in

        return data

"""
Entity Models
These entity models are what all EVE Online data models are attached to.
"""
from django_eveonline_connector.utilities.esi.universe import *
from django_eveonline_connector.utilities.static.universe import *

class EveEntity(models.Model):
    name = models.CharField(max_length=128)
    external_id = models.IntegerField(unique=True)

    def __str__(self):
        return self.name


class EveCharacter(EveEntity):
    corporation = models.ForeignKey(
        "EveCorporation", on_delete=models.SET_NULL, null=True)
    roles = models.ManyToManyField("EveCorporationRole", blank=True)
    token = models.OneToOneField(
        "EveToken", on_delete=models.SET_NULL, null=True)

    @staticmethod
    def get_primary_character(user):
        primary_character_obj = PrimaryEveCharacterAssociation.objects.filter(
            user=user).first()
        if primary_character_obj:
            return primary_character_obj.character
        else:
            return None 

    @property
    def avatar(self):
        return "https://%s/characters/%s/%s" % (
            app_config.EVEIMG_CONFIG['domain'],
            self.external_id,
            app_config.EVEIMG_CONFIG['portrait_mapping']['character'],
        )

    @staticmethod
    def create_from_external_id(external_id):
        response = EveClient.call(
            'get_characters_character_id', character_id=external_id)

        eve_character = EveCharacter(
            name=response.data['name'],
            external_id=external_id)

        eve_character.save()

        return eve_character

    def update_character_corporation(self, corporation_id=None):
        if not corporation_id:
            response = EveClient.call('post_characters_affiliation', characters=[self.external_id])
            corporation_id = response.data[0]['corporation_id']

        if EveCorporation.objects.filter(external_id=corporation_id).exists():
            self.corporation = EveCorporation.objects.get(external_id=corporation_id)
        else:
            self.corporation = EveCorporation.create_from_external_id(external_id=corporation_id)
        
        self.save() 


class EveCorporation(EveEntity):
    ceo = models.OneToOneField(
        "EveCharacter", blank=True, null=True, on_delete=models.SET_NULL)
    ticker = models.CharField(max_length=5)
    alliance = models.ForeignKey(
        "EveAlliance", blank=True, null=True, on_delete=models.SET_NULL)

    track_corporation = models.BooleanField(default=False)
    track_characters = models.BooleanField(default=False)

    @staticmethod
    def create_from_external_id(external_id):
        response = EveClient.call(
            'get_corporations_corporation_id', corporation_id=external_id)
        eve_corporation = EveCorporation(
            name=response.data['name'],
            ticker=response.data['ticker'],
            external_id=external_id)

        eve_corporation.save()

        return eve_corporation

class EveAlliance(EveEntity):
    executor = models.OneToOneField(
        "EveCorporation", blank=True, null=True, on_delete=models.SET_NULL)
    ticker = models.CharField(max_length=5)

    def save(self, *args, **kwargs):
        super(EveAlliance, self).save(*args, **kwargs)

    @staticmethod
    def create_from_external_id(external_id):
        response = EveClient.call(
            'get_alliances_alliance_id', alliance_id=external_id)

        eve_alliance = EveAlliance(
            name=response.data['name'],
            ticker=response.data['ticker'],
            external_id=external_id)

        eve_alliance.save()

        return eve_alliance


"""
Entity Data Models 
These models represent entity data. 
"""


class EveEntityData(models.Model):
    entity = models.ForeignKey(EveEntity, on_delete=models.CASCADE)

    @classmethod
    def create_from_esi_row(cls, data_row, entity_external_id, *args, **kwargs):
        """
        Wrapper function around self._create_from_esi_row, which is defined by the extending class. 
        Converts the data from a single ESI row (e.g jump_clones = []) into a Django Model and saves it.
        """
        try:
            db_object = cls._create_from_esi_row(
                data_row, entity_external_id, *args, **kwargs)
            db_object.entity = EveEntity.objects.get(
                external_id=entity_external_id)
            db_object.save()
        except Exception as e:
            message = (
                f"Failed to create {cls.__name__} from ESI data row: {data_row}.\n"
                f"Failed for entity {entity_external_id}\n"
            )
            logger.warning(message)
            logger.error(e, exc_info=True)

    @staticmethod
    def _create_from_esi_row(data_row, entity_external_id, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def create_from_esi_response(cls, data, entity_external_id, *args, **kwargs):
        """
        Pre-processes the response from an ESI Call and generated Django database objects using cls.create_from_esi_row()
        Wrapper around self._create_esi_response, which is defined by the extending class. 
        """
        try:
            processed_data = cls._create_from_esi_response(data, entity_external_id, *args, **kwargs)
        except Exception as e:
            logger.warning(
                "Failed to process ESI response for %s. Data: %s" % (cls.__name__, data))
            logger.exception(e)

    @staticmethod
    def _create_from_esi_response(data, entity_external_id, *args, **kwargs):
        raise NotImplementedError

    class Meta:
        abstract = True


class EveAsset(EveEntityData):
    # ESI Data
    is_blueprint_copy = models.BooleanField()
    is_singleton = models.BooleanField()
    item_id = models.BigIntegerField()
    location_flag = models.CharField(max_length=64)
    location_id = models.BigIntegerField()
    location_type = models.CharField(max_length=64)
    quantity = models.IntegerField()
    type_id = models.IntegerField()

    # Mapped Static Data
    group_id = models.IntegerField()
    category_id = models.IntegerField()

    # Our Conversions
    item_name = models.CharField(max_length=128)  # typeName
    item_type = models.CharField(max_length=32)  # categoryName
    location_name = models.CharField(max_length=128)

    def __str__(self):
        return "<%s :: %s - %s>" % (self.entity, self.item_name, self.location_name)

    @staticmethod
    def get_bad_asset_category_ids():
        """
        Assets have some very ugly categories that can't be resolved to locations. e.g POCOs
        It's recommended that you just filter out assets in these categories.
        """
        return app_config.ESI_BAD_ASSET_CATEGORIES

    @staticmethod
    def _create_from_esi_response(data, entity_external_id, *args, **kwargs):
        for row in data:
            EveAsset.create_from_esi_row(row, entity_external_id)

    @staticmethod
    def _create_from_esi_row(data_row, entity_external_id, *args, **kwargs):
        # Store what ESI returns
        asset = EveAsset(
            is_singleton=data_row['is_singleton'],
            item_id=data_row['item_id'],
            location_flag=data_row['location_flag'],
            location_id=data_row['location_id'],
            location_type=data_row['location_type'],
            quantity=data_row['quantity'],
            type_id=data_row['type_id'])

        # Handle optional ESI fields
        if 'is_blueprint_copy' in data_row:
            asset.is_blueprint_copy = True
        else:
            asset.is_blueprint_copy = False

        # Map useful static data
        asset.group_id = resolve_type_id_to_group_id(asset.type_id)
        asset.category_id = resolve_type_id_to_category_id(asset.type_id)

        # Use the static database to resolve EVE Model IDs
        asset.item_name = resolve_type_id_to_type_name(asset.type_id)
        asset.item_type = resolve_type_id_to_category_name(asset.type_id)

        if asset.category_id in EveAsset.get_bad_asset_category_ids():
            asset.location_name = "Unresolved Location"
            return asset # skip resolving of location

        # Location IDs suck to resolve, we must do different things for each type
        if asset.location_flag == "Hangar":
            asset.location_name = resolve_location_from_location_id_location_type(
                asset.location_id,
                asset.location_type,
                entity_external_id)

        return asset


class EveJumpClone(EveEntityData):
    # ESI Data
    location_id = models.BigIntegerField()
    location_type = models.CharField(max_length=64)
    jump_clone_id = models.IntegerField()

    # Our Conversions
    location = models.CharField(max_length=128)
    implants = models.TextField()

    @staticmethod
    def _create_from_esi_response(data, entity_external_id, *args, **kwargs):
        for row in data['jump_clones']:
            EveJumpClone.create_from_esi_row(row, entity_external_id)

    @staticmethod
    def _create_from_esi_row(data_row, entity_external_id, *args, **kwargs):
        """
        URL: /characters/{character_id}/clones/
        Expects a row from `jump_clones`
        """
        clone = EveJumpClone(
            location_id=data_row['location_id'],
            location_type=data_row['location_type'],
            jump_clone_id=data_row['jump_clone_id'],)

        logger.debug(data_row)
        clone.location = resolve_location_from_location_id_location_type(
            clone.location_id,
            clone.location_type,
            entity_external_id)

        implants = []
        for implant in data_row['implants']:
            implants.append(resolve_type_id_to_type_name(implant))

        clone.implants = ",".join(implants)
        return clone


class EveContact(EveEntityData):
    # ESI Data
    contact_id = models.IntegerField()
    is_blocked = models.BooleanField(default=False)
    is_watched = models.BooleanField(default=False)
    label_ids = models.TextField(blank=True, null=True)
    standing = models.FloatField()

    # Our Conversions
    contact_name = models.CharField(max_length=64)
    contact_type = models.CharField(max_length=32, choices=id_types)

    @property
    def evewho(self):
        return "https://%s/%s/%s" % (app_config.EVEWHO_CONFIG['domain'],
                                     self.contact_type,
                                     self.contact_id)

    @property
    def portrait(self):
        return "https://%s/%ss/%s" % (app_config.EVEIMG_CONFIG['domain'],
                                      self.contact_type,
                                      app_config.EVEIMG_CONFIG['portrait_mapping'][self.contact_type])

    @staticmethod
    def _create_from_esi_response(data, entity_external_id, *args, **kwargs):
        if len(data) == 0:
            return # no contacts
        # process user ids
        contact_ids_to_resolve = set()
        for contact in data:
            contact_ids_to_resolve.add(contact['contact_id'])
        contact_names = resolve_ids(list(contact_ids_to_resolve))
        for row in data:
            EveContact.create_from_esi_row(
                row, entity_external_id, contact_name=contact_names[row['contact_id']])

    @staticmethod
    def _create_from_esi_row(data_row, entity_external_id, *args, **kwargs):
        # Store what ESI returns
        contact = EveContact(
            contact_id=data_row['contact_id'],
            contact_type=data_row['contact_type'],
            standing=data_row['standing'],
            is_blocked=data_row.get('is_blocked', False),
            is_watched=data_row.get('is_watched', False),
        )

        # Handle optional ESI fields
        if 'label_ids' in data_row:
            label_ids = []
            for label_id in data_row['label_ids']:
                label_ids.append(str(label_id))
            if len(label_ids) >= 1:
                contact.label_ids = ",".join(label_ids)

        # Our Conversions
        if 'contact_name' not in kwargs:
            logger.warning(
                "Called without contact_name keyword argument. Performance decrease.")
            contact.contact_name = resolve_ids([ data_row['contact_id'] ])[
                contact.contact_id]
        else:
            contact.contact_name = kwargs.get('contact_name')
        return contact


class EveContract(EveEntityData):
    # ESI Data
    contract_id = models.BigIntegerField(unique=True)
    acceptor_id = models.BigIntegerField()
    assignee_id = models.BigIntegerField()

    buyout = models.FloatField(blank=True, null=True)
    collateral = models.FloatField(blank=True, null=True)
    price = models.FloatField(blank=True, null=True)
    reward = models.FloatField(blank=True, null=True)
    volume = models.FloatField(blank=True, null=True)

    date_accepted = models.DateTimeField(blank=True, null=True)
    date_completed = models.DateTimeField(blank=True, null=True)
    date_expired = models.DateTimeField()
    date_issued = models.DateTimeField()
    days_to_complete = models.IntegerField(blank=True, null=True)

    for_corporation = models.BooleanField()
    availability = models.CharField(max_length=32)
    issuer_corporation_id = models.BigIntegerField()
    issuer_id = models.BigIntegerField()

    start_location_id = models.BigIntegerField(blank=True, null=True)
    end_location_id = models.BigIntegerField(blank=True, null=True)
    status = models.CharField(max_length=32)
    title = models.TextField(blank=True, null=True)
    type = models.CharField(max_length=16)

    # Our Conversions
    acceptor_name = models.CharField(max_length=64)
    assignee_name = models.CharField(max_length=64)
    issuer_name = models.CharField(max_length=64)
    issuer_corporation_name = models.CharField(max_length=64)
    acceptor_type = models.CharField(max_length=32, choices=id_types)
    assignee_type = models.CharField(max_length=32, choices=id_types)
    issuer_type = models.CharField(max_length=64, choices=id_types)
    items = models.TextField(null=True)

    @staticmethod
    def _create_from_esi_response(data, entity_external_id, *args, **kwargs):
        ids_to_resolve = set()
        for contract in data:
            ids_to_resolve.add(contract['acceptor_id'])
            ids_to_resolve.add(contract['assignee_id'])
            ids_to_resolve.add(contract['issuer_id'])
            ids_to_resolve.add(contract['issuer_corporation_id'])
        resolved_ids = resolve_ids_with_types(ids_to_resolve)

        for row in data:
            EveContract.create_from_esi_row(
                row, entity_external_id, resolved_ids=resolved_ids, corporation=False)

    @staticmethod
    def _create_from_esi_row(data_row, entity_external_id, *args, **kwargs):
        """
        Fill an EveContract object according to the provided ESI data row.

        **Args**
            **data_row** (``dict``)
                The ESI data row that represents this object 
            **entity_external_id** (``int``)
                The External ID for the EveEntity

        **Kwargs**
            **resolved_ids** (``list``)
                Pre-resolved IDs for assignees, acceptors, etc. 
            **corporation** (``boolean``)
                If set, assumes we are resolving corporation contract items.

        **Returns**
            EveContract() instance
        """
        contract_id = data_row['contract_id']
        if EveContract.objects.filter(contract_id=contract_id).exists():
            contract = EveContract.objects.get(contract_id=contract_id)
            update = True 
        else:
            contract = EveContract()
            update = False 

        for key in data_row.keys():
            try:
                if type(data_row[key]) == pyswagger.primitives._time.Datetime:
                    data_row[key] = data_row[key].to_json()
                setattr(contract, str(key), data_row[key])
            except AttributeError:
                logger.error(f"Encountered unknown attribute {key} for EveContract")

        if update:
            return contract # the other fields will already be sorted out if we're doing an update

        if not 'resolved_ids' in kwargs:
            logger.warning(
                "Resolved IDs were not passed. Performance decrease.")
            ids_to_resolve = [contract.acceptor_id, contract.assignee_id,
                              contract.issuer_id, contract.issuer_corporation_id]
            resolved_ids = resolve_ids_with_types(ids_to_resolve)
            # todo resolve IDs
        else:
            resolved_ids = kwargs.get('resolved_ids')


        contracts_response = EveClient.call(
            'get_characters_character_id_contracts_contract_id_items',
            character_id=entity_external_id, contract_id=contract.contract_id)
    
        if not contracts_response:
            items = None 
        else:
            items = []
            for item in contracts_response.data:
                items.append("%s %s" % (
                    item['quantity'], resolve_type_id_to_type_name(item['type_id'])))
            contract.items = "\n".join(items)

        contract.acceptor_name = resolved_ids[contract.acceptor_id]['name']
        contract.assignee_name = resolved_ids[contract.assignee_id]['name']
        contract.issuer_name = resolved_ids[contract.issuer_id]['name']
        contract.acceptor_type = resolved_ids[contract.acceptor_id]['type']
        contract.assignee_type = resolved_ids[contract.assignee_id]['type']
        contract.issuer_type = resolved_ids[contract.issuer_id]['type']
        contract.issuer_corporation_name = resolved_ids[contract.issuer_corporation_id]

        return contract


class EveSkill(EveEntityData):
    # ESI Data
    active_skill_level = models.IntegerField()
    trained_skill_level = models.IntegerField()
    skill_id = models.IntegerField()
    skillpoints_in_skill = models.IntegerField()

    # Our Conversions
    skill_name = models.CharField(max_length=64)
    skill_group = models.CharField(max_length=64)

    class Meta:
        unique_together = ['entity', 'skill_name']

    @staticmethod
    def _create_from_esi_response(data, entity_external_id, *args, **kwargs):
        for row in data.skills:
            EveSkill.create_from_esi_row(row, entity_external_id)

    @staticmethod
    def _create_from_esi_row(data_row, entity_external_id, *args, **kwargs):
        skill = EveSkill(
            active_skill_level=data_row['active_skill_level'],
            trained_skill_level=data_row['trained_skill_level'],
            skill_id=data_row['skill_id'],
            skillpoints_in_skill=data_row['skillpoints_in_skill'],
        )

        skill.skill_name = resolve_type_id_to_type_name(data_row['skill_id'])
        skill.skill_group = resolve_type_id_to_group_name(data_row['skill_id'])

        return skill


class EveJournalEntry(EveEntityData):
    # ESI Data
    external_id = models.BigIntegerField(unique=True)
    amount = models.FloatField(blank=True, null=True)
    balance = models.FloatField(blank=True, null=True)
    tax = models.FloatField(blank=True, null=True)
    tax_receiver_id = models.IntegerField(blank=True, null=True)
    context_id = models.BigIntegerField(blank=True, null=True)
    context_id_type = models.CharField(max_length=32, blank=True, null=True)
    date = models.DateTimeField()
    description = models.TextField()
    reason = models.TextField()
    ref_type = models.CharField(max_length=128)
    first_party_id = models.IntegerField(blank=True, null=True)
    second_party_id = models.IntegerField(blank=True, null=True)

    # Our Conversions
    first_party_name = models.CharField(max_length=64)
    first_party_type = models.CharField(max_length=32, choices=id_types)
    second_party_name = models.CharField(max_length=64)
    second_party_type = models.CharField(max_length=32, choices=id_types)

    @staticmethod
    def _create_from_esi_response(data, entity_external_id, *args, **kwargs):
        ids_to_resolve = set()
        for row in data:
            ids_to_resolve.add(row['first_party_id'])
            ids_to_resolve.add(row['second_party_id'])
        resolved_ids = resolve_ids_with_types(ids_to_resolve)

        for row in data:
            if EveJournalEntry.objects.filter(external_id=row['id']).exists():
                continue
            EveJournalEntry.create_from_esi_row(
                row, entity_external_id, resolved_ids=resolved_ids)

    @staticmethod
    def _create_from_esi_row(data_row, entity_external_id, *args, **kwargs):
        entry = EveJournalEntry(
            external_id=data_row['id'],
            date=parse_datetime(data_row['date'].to_json()),
            description=data_row['description'],
            reason=data_row['description'],
            ref_type=data_row['ref_type'],
            amount=data_row.get('amount', None),
            balance=data_row.get('balance', None),
            tax=data_row.get('tax', None),
            tax_receiver_id=data_row.get('tax_receiver_id', None),
            context_id=data_row.get('context_id', None),
            context_id_type=data_row.get('context_id_type', None),
            first_party_id=data_row.get('first_party_id', None),
            second_party_id=data_row.get('second_party_id', None),
        )

        if 'resolved_ids' not in kwargs:
            logger.warning(
                "Called without Resolved IDs. Performance decrease.")
            ids_to_resolve = [entry.first_party_id, entry.second_party_id]
            resolved_ids = resolved_ids_with_types(ids_to_resolve)
        else:
            resolved_ids = kwargs.get('resolved_ids')

        entry.first_party_name = resolved_ids[entry.first_party_id]['name']
        entry.first_party_type = resolved_ids[entry.first_party_id]['type']
        entry.second_party_name = resolved_ids[entry.second_party_id]['name']
        entry.second_party_type = resolved_ids[entry.second_party_id]['type']

        return entry


class EveTransaction(EveEntityData):
    # ESI Data
    client_id = models.IntegerField()
    date = models.DateTimeField()
    is_buy = models.BooleanField()
    is_personal = models.BooleanField()
    journal_ref_id = models.BigIntegerField()
    location_id = models.BigIntegerField()
    quantity = models.IntegerField()
    transaction_id = models.BigIntegerField(unique=True)
    type_id = models.IntegerField()
    unit_price = models.FloatField()

    # Our Conversions
    client_name = models.CharField(max_length=64)
    client_type = models.CharField(max_length=64, choices=id_types)
    location_name = models.CharField(max_length=64)
    item_name = models.CharField(max_length=64)

    @property
    def get_journal_entry(self):
        if EveJournalEntry.objects.filter(external_id=self.journal_ref_id).exists():
            return EveJournalEntry.objects.get(external_id=self.journal_ref_id)
        return None

    @staticmethod
    def _create_from_esi_response(data, entity_external_id, *args, **kwargs):
        ids_to_resolve = set()
        for row in data:
            ids_to_resolve.add(row['client_id'])
        resolved_ids = resolve_ids_with_types(ids_to_resolve)

        for row in data:
            if EveTransaction.objects.filter(transaction_id=row['transaction_id']):
                continue
            EveTransaction.create_from_esi_row(
                row, entity_external_id, resolved_ids=resolved_ids)

    @staticmethod
    def _create_from_esi_row(data_row, entity_external_id, *args, **kwargs):
        transaction = EveTransaction(
            client_id=data_row['client_id'],
            date=parse_datetime(data_row['date'].to_json()),
            is_buy=data_row['is_buy'],
            is_personal=data_row['is_personal'],
            journal_ref_id=data_row['journal_ref_id'],
            location_id=data_row['location_id'],
            quantity=data_row['quantity'],
            transaction_id=data_row['transaction_id'],
            type_id=data_row['type_id'],
            unit_price=data_row['unit_price'],
        )

        if 'resolved_ids' not in kwargs:
            logger.warning(
                "Called without Resolved IDs. Performance decrease.")
            ids_to_resolve = [transaction.client_id]
            resolved_ids = resolved_ids_with_types(ids_to_resolve)
        else:
            resolved_ids = kwargs.get('resolved_ids')

        transaction.client_name = resolved_ids[transaction.client_id]['name']
        transaction.client_type = resolved_ids[transaction.client_id]['type']
        transaction.item_name = resolve_type_id_to_type_name(
            transaction.type_id)

        try:
            transaction.location_name = resolve_location_from_location_id_location_type(
                transaction.location_id, 'station', entity_external_id)
        except Exception as e:
            transaction.location_name = resolve_location_from_location_id_location_type(
                transaction.location_id, 'other', entity_external_id)
        
        if not transaction.location_name:
            transaction.location_name = "Unknown Location"

        return transaction

"""
Other EVE Models
"""
class EveStructure(EveEntityData):
    # Base ESI Data
    corporation_id = models.BigIntegerField()
    fuel_expires = models.DateTimeField(null=True, blank=True)
    next_reinforce_apply = models.DateTimeField(null=True, blank=True)
    next_reinforce_hour = models.BigIntegerField(null=True, blank=True)
    next_reinforce_weekday = models.BigIntegerField(null=True, blank=True)
    profile_id = models.BigIntegerField()
    reinforce_hour = models.BigIntegerField()
    reinforce_weekday = models.BigIntegerField(null=True, blank=True)
    services = models.TextField(null=True, blank=True)
    state = models.CharField(max_length=64)
    state_timer_end = models.DateTimeField(null=True, blank=True)
    state_timer_start = models.DateTimeField(null=True, blank=True)
    structure_id = models.BigIntegerField()
    system_id = models.BigIntegerField()
    type_id = models.BigIntegerField()
    unanchors_at = models.DateTimeField(null=True, blank=True)

    # Additional ESI Data
    owner_id = models.BigIntegerField()
    solar_system_id = models.BigIntegerField()
    type_id = models.BigIntegerField(null=True, blank=True)
    name = models.CharField(max_length=256)

    @property
    def fuel_expires_soon(self):
        if self.fuel_expires and ((self.fuel_expires - timezone.now()).days < 7):
            return True
        return False

    @property
    def reinforcement_time(self):
        return timezone.now().replace(hour=0, minute=0, second=0,
                                      microsecond=0) + datetime.timedelta(hours=self.reinforce_hour)

    def __str__(self):
        if not self.name:
            self.name = "Unknown Name"
        return f"{self.name}"


    @staticmethod
    def _create_from_esi_response(data, entity_external_id, *args, **kwargs):
        for row in data:
            EveStructure.create_from_esi_row(row, entity_external_id)

    @staticmethod
    def _create_from_esi_row(data_row, entity_external_id, *args, **kwargs):
        logger.info("Creating EVE Structure")
        if EveStructure.objects.filter(structure_id=data_row['structure_id']).exists():
            structure = EveStructure.objects.filter(structure_id=data_row['structure_id']).first()
        else:
            structure = EveStructure()
        for key in data_row.keys():
            try:
                if type(data_row[key]) == pyswagger.primitives._time.Datetime:
                    data_row[key] = data_row[key].to_json()
                setattr(structure, str(key), data_row[key])
            except AttributeError:
                logger.error(f"Encountered unknown attribute {key} for EveStructure")

        additional_info = get_structure(
            data_row['structure_id'], entity_external_id).data

        for key in additional_info.keys():
            try:
                setattr(structure, str(key), additional_info[key])
            except AttributeError:
                logger.error(
                    f"Encountered unknown attribute {key} for EveStructure")

        return structure

    class Meta:
        permissions = [
            ('bypass_corporation_view_requirements',
             "Can view a corporation's structures regardless of current membership"),
        ]


"""
Static EVE Models
These represent static EVE models that we need associations for.
"""
class EveCorporationRole(models.Model):
    codename = models.CharField(max_length=64, unique=True)
    name = models.CharField(max_length=64, unique=True)

    def __str__(self):
        return self.name 
            
"""
EVE Connector Models
Models for EVE Connector functionality 
"""
class PrimaryEveCharacterAssociation(models.Model):
    user = models.OneToOneField(
        User, on_delete=models.CASCADE, related_name="primary_evecharacter")
    character = models.OneToOneField(
        EveCharacter, on_delete=models.CASCADE, related_name="primary_to")

    def __str__(self):
        return "<%s:%s>" % (self.user, self.character)

    def get_character(self):
        return self.character

    def set_character(self, character):
        self.character = character
        self.save()

    def get_user(self):
        return self.user

    def set_user(self):
        self.user = user
        self.save()

    def get_related_characters(self):
        return EveCharacter.objects.filter(token__in=EveToken.objects.filter(user=self.user)).exclude(external_id=self.character.external_id)

class EveCharacterInfo(models.Model):
    # TODO: Merge this into EveCharacter
    character = models.OneToOneField(EveCharacter, on_delete=models.CASCADE, related_name="info")
    
    # game info
    skill_points = models.IntegerField(blank=True, null=True)
    net_worth = models.FloatField(blank=True, null=True)
    
    # update statistics
    assets_last_updated = models.DateTimeField(blank=True, null=True)
    jump_clones_last_updated = models.DateTimeField(blank=True, null=True)
    contacts_last_updated = models.DateTimeField(blank=True, null=True)
    contracts_last_updated = models.DateTimeField(blank=True, null=True)
    skills_last_updated = models.DateTimeField(blank=True, null=True)
    journal_last_updated = models.DateTimeField(blank=True, null=True)
    transactions_last_updated = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return self.character.name

class EveGroupRule(models.Model):
    group = models.OneToOneField(Group, on_delete=models.CASCADE)
    include_alts = models.BooleanField(default=True)
    
    # qualifiers 
    characters = models.ManyToManyField("EveCharacter", blank=True)
    corporations = models.ManyToManyField("EveCorporation", blank=True)
    alliances = models.ManyToManyField("EveAlliance", blank=True)
    roles = models.ManyToManyField("EveCorporationRole", blank=True)

    @property
    def current_user_list(self):
        return User.objects.filter(group__contains=self.group)

    @property
    def valid_user_list(self):
        django_filter = Q()
        if self.characters.all():
            django_filter &= Q(evecharacter__in=self.characters.all())
        if self.corporations.all():
            django_filter &= Q(
                evecharacter__corporation__in=self.corporations.all())
        if self.alliances.all():
            django_filter &= Q(
                evecharacter__corporation__alliance__in=self.alliances.all())
        if self.roles.all():
            django_filter &= Q(evecharacter__roles__in=self.roles.all())
        if not self.include_alts:
            django_filter &= ~Q(evecharacter__primary_to=None)
        if django_filter != Q():
            return User.objects.filter(eve_tokens__in=EveToken.objects.filter(django_filter))
        else:
            return User.objects.none()

    @property 
    def missing_user_list(self):
        return self.valid_user_list.filter(~Q(groups=self.group))

    @property 
    def invalid_user_list(self):
        django_filter = Q()
        if self.characters.all():
            django_filter &= Q(evecharacter__in=self.characters.all())
        if self.corporations.all():
            django_filter &= Q(
                evecharacter__corporation__in=self.corporations.all())
        if self.alliances.all():
            django_filter &= Q(
                evecharacter__corporation__alliance__in=self.alliances.all())
        if self.roles.all():
            django_filter &= Q(evecharacter__roles__in=self.roles.all())
        if not self.include_alts:
            django_filter &= ~Q(evecharacter__primary_to=None)
        if django_filter != Q():
            return User.objects.filter(groups=self.group).exclude(eve_tokens__in=EveToken.objects.filter(django_filter))
        else:
            return User.objects.none()



# Generated by Django 2.2.13 on 2020-10-04 01:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0035_auto_20201004_0130'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='evestructure',
            options={'permissions': [('bypass_corporation_view_requirements', "Can view a corporation's structures regardless of current membership")]},
        ),
    ]



from django.db import connections
from django_eveonline_connector.utilities.esi.universe import *
from django_eveonline_connector.exceptions import EveDataResolutionError
from django.db.utils import ConnectionDoesNotExist, OperationalError
import django.db.utils
import logging

logger = logging.getLogger(__name__)

def query_static_database(query, raise_exception=False, fetchall=False):
    """
    Executes the SQL query on the 'eve_static' database.
    If the value is not found, returns None. 
    """
    try:
        with connections['eve_static'].cursor() as cursor:
            cursor.execute(query)
            if fetchall:
                result = cursor.fetchall()
            else:
                result = cursor.fetchone()
        if result:
            if fetchall:
                return result
            return result[0]
    except django.db.utils.DatabaseError as e:
        logger.error("EVE Static Database is not properly configured, or ran into an error.")
        if raise_exception:
            raise(e)
        else:
            logger.exception(e)
        
    return None 

def resolve_type_id_to_type_name(type_id, raise_exception=True):
    """
    Resolves an EVE Online type_id to type_name. Falls back to ESI if not found in static database.

    Returns None if not found.
    """
    query = "select typeName from invTypes where typeID = %s" % type_id
    type_name = query_static_database(query)
    
    if type_name:
        return type_name 
    

    logger.info("Resolving type_id (%s) using ESI" % type_id)
    type_name = get_type_id(type_id)

    if (not type_name or 'name' not in type_name) and raise_exception:
        raise EveDataResolutionError(f"Failed to resolve type_id ({type_id}) to type_name")

    return type_name['name']

def resolve_type_name_to_type_id(type_name, raise_exception=True):
    """
    Resolve type_name to type_id.
    WARNING: Prone to SQL injection, do not expose to user.
    """
    type_name = type_name.replace("'", "''").rstrip()
    query = "select typeID from invTypes where typeName = '%s'" % type_name
    type_id = query_static_database(query)

    if type_id:
        return type_id         

    logger.info("Resolving type_id from type_name (%s) using ESI" % type_name)
    response = resolve_names(type_name)
    if 'inventory_types' in response and response['inventory_types'] and 'id' in response['inventory_types'][0]:
        type_id = response['inventory_types'][0]['id']
    
    if not type_id and raise_exception:
        raise EveDataResolutionError(f"Error resolving EVE type_name({type_name}) to type_id")

    return type_id 

def resolve_type_id_to_group_id(type_id, raise_exception=True):
    query = "select groupID from invTypes where typeID = %s" % type_id
    group_id = query_static_database(query)
    
    if group_id:
        return group_id 

    logger.info("Resolving group_id from type_id (%s) using ESI" % type_id)

    response = get_type_id(type_id)
    if 'group_id' in response:
        group_id = response['group_id']

    if not group_id and raise_exception:
        raise EveDataResolutionError(
            f"Error resolving EVE type_id({type_id}) to group_id")

    return group_id 


def resolve_type_id_to_category_name(type_id, raise_exception=True):
    category_id = resolve_type_id_to_category_id(type_id)
    category_name = resolve_category_id_to_category_name(category_id)
    if not category_name and raise_exception:
        raise EveDataResolutionError(
            f"Failed to resolve type_id({type_id}) to category name")
    return category_name

def resolve_group_id_to_group_name(group_id, raise_exception=True):
    query = "select groupName from invGroups where groupID = %s" % group_id
    group_name = query_static_database(query)

    if group_name:
        return group_name

    logger.info("Resolving group_name from group_id (%s) using ESI" % group_id)
    response = get_group_id(group_id)
    if 'name' in response:
        group_name = response['name']

    if not group_name and raise_exception:
        raise EveDataResolutionError(
            f"Error resolving EVE type_id({group_id}) to group_id")

    return group_name 


def resolve_type_id_to_group_name(type_id):
    group_id = resolve_type_id_to_group_id(type_id)
    return resolve_group_id_to_group_name(group_id)


def resolve_group_id_to_category_id(group_id, raise_exception=True):
    query = "select categoryID from invGroups where groupID = %s" % group_id
    category_id = query_static_database(query)

    if category_id:
        return category_id

    logger.info("Resolving category_id from group_id (%s) using ESI" % group_id)

    response = get_group_id(group_id)
    if 'category_id' in response:
        category_id = response['category_id']
    
    if not category_id and raise_exception:
        raise EveDataResolutionError(
            f"Error resolving category_id from group_id ({group_id})")
    
    return category_id

def resolve_category_id_to_category_name(category_id, raise_exception=True):
    query = "select categoryName from invCategories where categoryID = %s" % category_id
    category_name = query_static_database(query)

    if category_name:
        return category_name 

    logger.info("Resolving category_name from category_id (%s) using ESI" % category_id)
    
    response = get_category_id(category_id)
    if 'name' in response:
        category_name = response['name']

    if not category_name and raise_exception:
        raise EveDataResolutionError(
            f"Error resolving category_name from category_id ({category_id})")

    return category_name

def resolve_type_id_to_category_id(type_id):
    group_id = resolve_type_id_to_group_id(type_id)
    category_id = resolve_group_id_to_category_id(group_id)
    return int(category_id)


def resolve_location_id_to_station(location_id, raise_exception=True):
    if location_id < 60000000 or location_id > 64000000:
        raise EveDataResolutionError('Attempted to resolve a station outside of the possible range')
    
    query = "select stationName from staStations where stationID = %s" % location_id
    station_name = query_static_database(query)

    if station_name:
        return station_name

    if str(location_id) in cache:
        return cache.get(str(location_id))

    logger.info(
        "Resolving location_id (%s) to station using ESI" % location_id)
    response = get_station_id(location_id)

    if 'name' in response:
        station_name = response['name']
        cache.set(str(location_id), station_name)

    if not station_name and raise_exception:
        raise EveDataResolutionError(f"Failed to resolve location_id({location_id} to station")
    elif not station_name:
        return "Unknown Station"
    
    return station_name 

def resolve_location_id(location_id, token_entity_id):
    location_name = resolve_location_id_to_station(location_id)
    if location_name:
        return location_name

    location_name = get_structure_id(location_id, token_entity_id)
    if location_name:
        return location_name

    return "Unknown Location"
    
def resolve_location_from_location_id_location_type(location_id, location_type, token_entity_id):
    location = f"Unknown Location ({location_id})"
    logger.debug("Resolving location_id (%s) of location_type(%s) to location_name" %
        (location_id, location_type))
    try:
        if location_type == 'station':
            location = resolve_location_id_to_station(location_id)
        else:
            location = get_structure_id(location_id, token_entity_id)
            
    except Exception as e:
        raise EveDataResolutionError(
            f"Failed to resolve location_id ({location_id}) of location_type ({location_type}). Reason: {str(e)}")

    return location

def get_type_id_prerq_skill_ids(type_id):
    query = """SELECT 
        i.typeID         as itemID, 
        ip.typeID        as prerqSkillID,
        dtal.valueInt    as prerqSkillLevelInt,
        dtal.valueFloat  as prerqSkillLevelFloat
    FROM invGroups g
    LEFT JOIN invTypes i 
        ON i.groupID = g.groupID
    LEFT JOIN dgmTypeAttributes dta
        ON dta.typeID = i.typeID AND
            dta.attributeID IN (182, 183, 184, 1285, 1289, 1290)
    LEFT JOIN dgmTypeAttributes dtal 
        ON dtal.typeID = dta.typeID AND 
        (
            (dtal.attributeID = 277 AND dta.attributeID = 182) OR
            (dtal.attributeID = 278 AND dta.attributeID = 183) OR
            (dtal.attributeID = 279 AND dta.attributeID = 184) OR
            (dtal.attributeID = 1286 AND dta.attributeID = 1285) OR
            (dtal.attributeID = 1287 AND dta.attributeID = 1289) OR
            (dtal.attributeID = 1288 AND dta.attributeID = 1290)
        )
    JOIN invTypes ip 
        ON ip.typeID = dta.valueInt OR
            ip.typeID = dta.valueFloat

    WHERE i.typeID=%s
        AND i.published = 1
        AND g.categoryID NOT IN (0,1,2,3,25)
    ORDER BY g.groupName DESC
    """ % str(type_id)

    try:
        with connections['eve_static'].cursor() as cursor:
            cursor.execute(query)
            result = cursor.fetchall()
    except django.db.utils.DatabaseError as e:
        raise EveDataResolutionError(
            f"EVE Static Database is not properly configured, or ran into an error. {e}")
    return result


def get_prerequisite_skills(type_id):
    skill_to_resolve = type_id.pop(0)
    prereqs = get_type_id_prerq_skill_ids(skill_to_resolve[1])
    skill_to_save = { 
        'name': resolve_type_id_to_type_name(skill_to_resolve[1]),
        'type_id': skill_to_resolve[1],
        'level': skill_to_resolve[3],
    }
    if prereqs or type_id:
        return [skill_to_save] + get_prerequisite_skills(type_id + prereqs)
    else:
        return [skill_to_save]



from django.apps import AppConfig
from django.apps import apps
import os 

class DjangoEveOnlineConnectorConfig(AppConfig):
    name = "django_eveonline_connector"
    verbose_name = "EVE Online Connector"
    url_slug = 'eveonline'
    package_name = __import__(name).__package_name__
    version = __import__(name).__version__
    
    ESI_BASE_URL=os.environ.get('ESI_BASE_URL', "")
    ESI_CLIENT_ID=os.environ.get('ESI_CLIENT_ID', "")
    ESI_SECRET_KEY=os.environ.get('ESI_SECRET_KEY', "")
    ESI_CALLBACK_URL=os.environ.get('ESI_CALLBACK_URL', "")

    # DO NOT EDIT 
    ESI_BAD_ASSET_CATEGORIES=[42, 43]

    EVEWHO_CONFIG = {
        'domain': 'evewho.com'
    }

    EVEIMG_CONFIG = {
        'domain': 'images.evetech.net',
        'portrait_mapping': {
            'character': 'portrait',
            'corporation': 'logo',
            'alliance': 'logo',
        }
    }
    
    def ready(self):
        from django_eveonline_connector.models import EveClient 
        from django.db.utils import OperationalError

        # bind to krypted application
        if apps.is_installed('packagebinder'):
            from packagebinder.exceptions import BindException
            try:
                bind = apps.get_app_config('packagebinder').get_bind_object(
                    self.package_name, self.version)
                # Required Task Bindings
                bind.add_required_task(
                    name="EVE: Update Tokens",
                    task="django_eveonline_connector.tasks.update_tokens",
                    interval=1,
                    interval_period="days",
                )
                bind.add_required_task(
                    name="EVE: Update Characters",
                    task="django_eveonline_connector.tasks.update_characters",
                    interval=1,
                    interval_period="days",
                )
                bind.add_required_task(
                    name="EVE: Update Corporations",
                    task="django_eveonline_connector.tasks.update_corporations",
                    interval=1,
                    interval_period="days",
                )
                bind.add_required_task(
                    name="EVE: Update Affilitions",
                    task="django_eveonline_connector.tasks.update_tokens",
                    interval=5,
                    interval_period="minutes",
                )
                
                # Optional Task Bindings
                bind.add_optional_task(
                    name="EVE: Update Structures",
                    task="django_eveonline_connector.tasks.update_structures",
                    interval=1,
                    interval_period="days",
                )
                bind.add_optional_task(
                    name="EVE: Update Character Corporation Roles",
                    task="django_eveonline_connector.tasks.update_character_roles",
                    interval=1,
                    interval_period="days",
                )
                bind.add_optional_task(
                    name="EVE: Assign Eve Groups",
                    task="django_eveonline_connector.tasks.assign_eve_groups",
                    interval=30,
                    interval_period="minutes",
                )
                bind.save()
            except BindException as e:
                print(e)
                return 
            except Exception as e:
                print(e)
                return



# Generated by Django 2.2.10 on 2020-02-25 21:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0012_auto_20200225_1921'),
    ]

    operations = [
        migrations.AddField(
            model_name='evetokentype',
            name='description',
            field=models.CharField(default='Empty', max_length=64),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2.10 on 2020-09-28 16:20

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0030_auto_20200928_0013'),
    ]

    operations = [
        migrations.CreateModel(
            name='EveCharacterInfo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('skill_points', models.IntegerField(blank=True, null=True)),
                ('net_worth', models.FloatField(blank=True, null=True)),
                ('assets_last_updated', models.DateTimeField(blank=True, null=True)),
                ('jump_clones_last_updated', models.DateTimeField(blank=True, null=True)),
                ('contacts_last_updated', models.DateTimeField(blank=True, null=True)),
                ('contracts_last_updated', models.DateTimeField(blank=True, null=True)),
                ('skills_last_updated', models.DateTimeField(blank=True, null=True)),
                ('journal_last_updated', models.DateTimeField(blank=True, null=True)),
                ('transactions_last_updated', models.DateTimeField(blank=True, null=True)),
                ('character', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='info', to='django_eveonline_connector.EveCharacter')),
            ],
        ),
    ]



# Generated by Django 2.2.4 on 2019-10-03 20:08

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0005_auto_20191003_1947'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='evetoken',
            unique_together=set(),
        ),
    ]



# Generated by Django 2.2.4 on 2019-08-20 19:55

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='EveEntity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('portrait', models.URLField(blank=True, max_length=255, null=True)),
                ('external_id', models.IntegerField(unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='EveScope',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='EveAlliance',
            fields=[
                ('eveentity_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='django_eveonline_connector.EveEntity')),
                ('ticker', models.CharField(max_length=5)),
            ],
            bases=('django_eveonline_connector.eveentity',),
        ),
        migrations.CreateModel(
            name='EveCharacter',
            fields=[
                ('eveentity_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='django_eveonline_connector.EveEntity')),
            ],
            bases=('django_eveonline_connector.eveentity',),
        ),
        migrations.CreateModel(
            name='EveWalletTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('client_id', models.IntegerField()),
                ('date', models.DateTimeField()),
                ('is_buy', models.BooleanField()),
                ('is_personal', models.BooleanField()),
                ('journal_ref_id', models.IntegerField()),
                ('location_id', models.IntegerField()),
                ('quantity', models.IntegerField()),
                ('transcation_id', models.IntegerField()),
                ('type_id', models.IntegerField()),
                ('type_name', models.CharField(max_length=32)),
                ('unit_price', models.FloatField()),
                ('entity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_eveonline_connector.EveEntity')),
            ],
        ),
        migrations.CreateModel(
            name='EveClient',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('esi_base_url', models.URLField(default='https://esi.evetech.net/latest/swagger.json?datasource=tranquility')),
                ('esi_callback_url', models.URLField()),
                ('esi_sso_url', models.URLField(editable=False)),
                ('esi_client_id', models.CharField(max_length=255)),
                ('esi_secret_key', models.CharField(max_length=255)),
                ('esi_scopes', models.ManyToManyField(blank=True, to='django_eveonline_connector.EveScope')),
            ],
        ),
        migrations.CreateModel(
            name='EveToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('access_token', models.TextField()),
                ('refresh_token', models.TextField()),
                ('expires_in', models.IntegerField(default=0)),
                ('expiry', models.DateTimeField(auto_now_add=True)),
                ('scopes', models.ManyToManyField(blank=True, to='django_eveonline_connector.EveScope')),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('character', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='eve_token', to='django_eveonline_connector.EveCharacter')),
            ],
        ),
        migrations.CreateModel(
            name='EveCorporation',
            fields=[
                ('eveentity_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='django_eveonline_connector.EveEntity')),
                ('ticker', models.CharField(max_length=5)),
                ('alliance', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_eveonline_connector.EveAlliance')),
                ('ceo', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_eveonline_connector.EveCharacter')),
            ],
            bases=('django_eveonline_connector.eveentity',),
        ),
        migrations.AddField(
            model_name='evecharacter',
            name='corporation',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_eveonline_connector.EveCorporation'),
        ),
        migrations.AddField(
            model_name='evealliance',
            name='executor',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_eveonline_connector.EveCorporation'),
        ),
    ]



# Generated by Django 2.2.10 on 2020-03-08 22:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0024_auto_20200308_2049'),
    ]

    operations = [
        migrations.AlterField(
            model_name='evecontact',
            name='label_ids',
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 2.2.8 on 2019-12-11 15:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0009_auto_20191210_1538'),
    ]

    operations = [
        migrations.AlterField(
            model_name='evecharacter',
            name='token',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='django_eveonline_connector.EveToken'),
        ),
    ]



# Generated by Django 2.2.10 on 2020-09-28 00:13

from django.db import migrations, models
import django_eveonline_connector.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0029_load_required_scopes'),
    ]

    operations = [
        migrations.AddField(
            model_name='evetoken',
            name='invalidated',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='evetoken',
            name='requested_scopes',
            field=models.ManyToManyField(blank=True, default=django_eveonline_connector.models.EveScope.get_required, related_name='requested_scopes', to='django_eveonline_connector.EveScope'),
        ),
        migrations.AlterField(
            model_name='evescope',
            name='name',
            field=models.CharField(max_length=128, unique=True),
        ),
    ]



# Generated by Django 2.2.10 on 2020-09-30 18:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0032_auto_20200929_1545'),
    ]

    operations = [
        migrations.AddField(
            model_name='evecorporation',
            name='track_characters',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='evecorporation',
            name='track_corporation',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-26 17:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0015_merge_20200226_1503'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='eveclient',
            options={'verbose_name': 'Eve Settings', 'verbose_name_plural': 'Eve Settings'},
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-25 19:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0011_auto_20191211_1624'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='eveclient',
            name='esi_sso_url',
        ),
        migrations.CreateModel(
            name='EveTokenType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=32, unique=True)),
                ('esi_sso_url', models.URLField(editable=False, max_length=2056)),
                ('scopes', models.ManyToManyField(blank=True, to='django_eveonline_connector.EveScope')),
            ],
        ),
    ]



# Generated by Django 2.2.13 on 2020-10-18 16:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0036_auto_20201004_0148'),
    ]

    operations = [
        migrations.AlterField(
            model_name='evecontract',
            name='acceptor_id',
            field=models.BigIntegerField(),
        ),
        migrations.AlterField(
            model_name='evecontract',
            name='assignee_id',
            field=models.BigIntegerField(),
        ),
        migrations.AlterField(
            model_name='evecontract',
            name='contract_id',
            field=models.BigIntegerField(unique=True),
        ),
        migrations.AlterField(
            model_name='evecontract',
            name='issuer_corporation_id',
            field=models.BigIntegerField(),
        ),
        migrations.AlterField(
            model_name='evecontract',
            name='issuer_id',
            field=models.BigIntegerField(),
        ),
    ]



# Generated by Django 2.2.13 on 2020-10-04 01:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0034_evestructure'),
    ]

    operations = [
        migrations.AlterField(
            model_name='evecontract',
            name='items',
            field=models.TextField(null=True),
        ),
    ]



from django.shortcuts import render, redirect
from django_eveonline_connector.models import EveClient, EveToken, EveCharacter, EveScope, EveCorporation, PrimaryEveCharacterAssociation
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django_eveonline_connector.tasks import update_character

import logging
logger = logging.getLogger(__name__)

"""
SSO Views
"""
@login_required
def sso_callback(request):
    code = request.GET.get('code', None)
    eve_client = EveClient.get_instance()

    # verify token
    esi_security = EveClient.get_esi_security()
    esi_token = esi_security.auth(code)
    esi_character = esi_security.verify()

    # create new token
    new_token = EveToken.objects.get_or_create(
        access_token=esi_token['access_token'],
        refresh_token=esi_token['refresh_token'],
        expires_in=esi_token['expires_in'],
        user=request.user
    )[0]

    # set scopes M2M
    scopes = EveScope.objects.filter(name__in=esi_character['scp'])
    if scopes.count() != len(esi_character['scp']):
        logger.error(
            f"Whoa there. Somehow we added a scope we don't know about. Pass this to Krypted Developers: \n ${esi_character['scp']}")
    new_token.scopes.set(scopes)

    # find or create character
    if EveCharacter.objects.filter(external_id=esi_character['sub'].split(":")[-1]).exists():
        character = EveCharacter.objects.get(
            external_id=esi_character['sub'].split(":")[-1])
        if character.token:
            old_token = character.token
            old_token.delete()
        character.token = new_token
        character.save()
    else:
        character = EveCharacter.objects.create(
            external_id=esi_character['sub'].split(":")[-1],
            name=esi_character['name'],
            token=new_token,
        )

    # if no primary user, set 
    if not PrimaryEveCharacterAssociation(user=request.user):
        PrimaryEveCharacterAssociation.objects.create(
            user=request.user,
            character=character
        )

    update_character.apply_async(args=[character.external_id])

    return redirect('app-dashboard')  # TODO: Redirect to EVE Character view

@login_required
def add_sso_token(request):
    return redirect(EveClient.get_instance().get_sso_url())

@login_required
def update_sso_token(request, token_id):
    eve_token = EveToken.objects.get(pk=token_id)
    return redirect(EveClient.get_instance().get_sso_url(
        EveScope.convert_to_list(eve_token.requested_scopes.all())
        ))

@login_required
def remove_sso_token(request, pk):
    eve_token = EveToken.objects.get(pk=pk)
    if request.user == eve_token.user:
        eve_token.delete()
        messages.success(request, "Successfully deleted EVE Online character")
    else:
        messages.error(request, "You cannot delete someone elses token.")
    return redirect("/")



# Generated by Django 2.2.13 on 2020-10-25 18:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0037_auto_20201018_1628'),
    ]

    operations = [
        migrations.AlterField(
            model_name='evestructure',
            name='fuel_expires',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='evestructure',
            name='next_reinforce_apply',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='evestructure',
            name='next_reinforce_hour',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='evestructure',
            name='next_reinforce_weekday',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='evestructure',
            name='reinforce_weekday',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='evestructure',
            name='services',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='evestructure',
            name='state_timer_end',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='evestructure',
            name='state_timer_start',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='evestructure',
            name='type_id',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='evestructure',
            name='unanchors_at',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



# Generated by Django 2.2.13 on 2020-09-20 22:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0027_auto_20200920_2150'),
    ]

    operations = [
        migrations.AlterField(
            model_name='evescope',
            name='required',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.4 on 2019-10-03 19:47

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0004_remove_eveclient_esi_scopes'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='eveentity',
            name='portrait',
        ),
        migrations.RemoveField(
            model_name='evetoken',
            name='character',
        ),
        migrations.AddField(
            model_name='evecharacter',
            name='token',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_eveonline_connector.EveToken'),
        ),
        migrations.AlterField(
            model_name='evetoken',
            name='user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='eve_tokens', to=settings.AUTH_USER_MODEL),
        ),
        migrations.DeleteModel(
            name='EveWalletTransaction',
        ),
    ]



# Generated by Django 2.2.4 on 2019-10-02 18:43

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0003_auto_20190903_2005'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='eveclient',
            name='esi_scopes',
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-27 19:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0018_remove_evetoken_primary'),
    ]

    operations = [
        migrations.CreateModel(
            name='EveAsset',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_blueprint_copy', models.BooleanField()),
                ('is_singleton', models.BooleanField()),
                ('item_id', models.IntegerField()),
                ('location_flag', models.CharField(max_length=64)),
                ('location_id', models.IntegerField()),
                ('location_type', models.CharField(max_length=64)),
                ('type_id', models.IntegerField()),
                ('item', models.CharField(max_length=128)),
                ('item_type', models.CharField(max_length=32)),
                ('location', models.CharField(max_length=128)),
                ('quantity', models.IntegerField()),
                ('entity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='assets', to='django_eveonline_connector.EveEntity')),
            ],
        ),
    ]



from django.contrib import admin, messages
from django.conf import settings 
from django.apps import apps 
from django import forms
from django_eveonline_connector.models import *
from django.urls import reverse
from django.utils.safestring import mark_safe


app_config = apps.get_app_config('django_eveonline_connector')
if app_config.ESI_SECRET_KEY and app_config.ESI_CLIENT_ID and app_config.ESI_CALLBACK_URL and app_config.ESI_BASE_URL:
    pass 
else:
    if apps.is_installed('django_singleton_admin'):
        # Highly Recommended: https://github.com/porowns/django-singleton-admin
        from django_singleton_admin.admin import DjangoSingletonModelAdmin
        @admin.register(EveClient) 
        class EveClientAdmin(DjangoSingletonModelAdmin):
            fieldsets = (
                ('General Settings', {
                    'fields': ('esi_callback_url', 'esi_client_id', 'esi_secret_key', )
                }),
                ('Advanced Settings', {
                    'classes': ('collapse', 'open'),
                    'fields': ('esi_base_url',)
                }),
            )
    else:
        admin.site.register(EveClient)

@admin.register(EveGroupRule)
class EveGroupRuleAdmin(admin.ModelAdmin):
    list_display = ('group',)
    filter_horizontal = ('characters', 'corporations', 'alliances', 'roles')

@admin.register(EveScope)
class EveScopeAdmin(admin.ModelAdmin):
    list_display = ('name', 'required')

@admin.register(EveCorporation)
class EveCorporationAdmin(admin.ModelAdmin):
    list_display = ('name', 'ceo', 'track_corporation', 'track_characters')
    search_fields = ('name', 'ceo__name')

@admin.register(EveCharacter)
class EveCharacterAdmin(admin.ModelAdmin):
    list_display = ('name', 'get_corporation', 'get_user')
    search_fields = ('name', )

    @mark_safe
    def get_corporation(self, obj):
        if obj.corporation:
            link = reverse("admin:django_eveonline_connector_evecorporation_change", args=[
                           obj.corporation.pk])
            return u'<a href="%s">%s</a>' % (link, obj.corporation.name)
        else:
            return None 
    get_corporation.allow_tags=True
    get_corporation.short_description = "Corporation"

    @mark_safe 
    def get_user(self, obj):
        if obj.token: 
            link = reverse("admin:auth_user_change", args=[obj.token.user.pk])
            return u'<a href="%s">%s</a>' % (link, obj.token.user.username)
        else:
            return None 
    get_user.short_description = "User"


# Generated by Django 2.2.10 on 2020-02-26 15:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0013_evetokentype_description'),
        ('django_eveonline_connector', '0014_load_intial_data'),
    ]

    operations = [
    ]



from django.http import HttpResponse
from django.shortcuts import render, redirect
from django_eveonline_connector.models import EveCharacter, EveSkill, PrimaryEveCharacterAssociation
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django_eveonline_connector.tasks import *
from django.db.models import Count
from django_eveonline_connector.utilities.esi.universe import resolve_names

import logging
logger = logging.getLogger(__name__)


@login_required
def set_primary_character(request, character_id):
    character = EveCharacter.objects.get(external_id=character_id)
    if not character.token:
        messages.warning(request, "You may not set a primary character that does not have a token.")
        return redirect("/")
    elif not character.token.user == request.user:
        messages.warning(request, "You may not set a primary character that you do not own.")
        return redirect("/")
    
    if PrimaryEveCharacterAssociation.objects.filter(user=request.user).exists():
        assoc = PrimaryEveCharacterAssociation.objects.get(user=request.user)
        assoc.character = character 
        assoc.save()
    else:
        PrimaryEveCharacterAssociation.objects.create(
            user=request.user,
            character=character
        )

    messages.success(request, "Successfully updated primary character to %s" % character)
    return redirect("/")


@login_required
def select_primary_character(request):
    return render(request, 'django_eveonline_connector/adminlte/select_primary_character.html', context={
        'characters': EveCharacter.objects.filter(token__user=request.user),
    })

@login_required
def select_character(request, required_scopes, next):
    return redirect(next, character_id)

@login_required
def refresh_character_public(request, external_id):
    character = EveCharacter.objects.get(external_id=external_id)
    character.token.refresh()
    character.update_character_corporation()
    messages.success(request, "Character successfully updated")
    return redirect("/")

@login_required
@permission_required('django_eveonline_connector.change_evecharacter', raise_exception=True)
def refresh_character(request, external_id):
    if 'fields' not in request.GET:
       return HttpResponse(status=400)
    scopes = request.GET['fields'].split(",")
    
    if 'assets' in scopes:
        update_character_assets.apply_async(args=[external_id])
    if 'jumpclones' in scopes:
        update_character_jumpclones.apply_async(args=[external_id])
    if 'contacts' in scopes:
        update_character_contacts.apply_async(args=[external_id])
    if 'contracts' in scopes:
        update_character_contracts.apply_async(args=[external_id])
    if 'skills' in scopes:
        update_character_skills.apply_async(args=[external_id])
    if 'journal' in scopes:
        update_character_journal.apply_async(args=[external_id])
    if 'transactions' in scopes:
        update_character_transactions.apply_async(args=[external_id])
        
    messages.info(request, f"Jobs queued to update: {scopes}")
    return redirect('django-eveonline-connector-view-character', external_id)

@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
def list_characters(request):
    return render(request, 'django_eveonline_connector/adminlte/characters/list_characters.html', context={
        'characters': EveCharacter.objects.all()
    })


@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
def view_character(request, external_id):
    context = {}
    context['character'] = EveCharacter.objects.get(external_id=external_id)
    context['message'] = "A String"
    return render(request, 'django_eveonline_connector/adminlte/characters/view_character.html', context)


@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
@permission_required('django_eveonline_connector.view_eveasset', raise_exception=True)
def view_character_assets(request, external_id):
    character = EveCharacter.objects.get(external_id=external_id)
    asset_summary = []
    asset_locations = EveAsset.objects.filter(entity__external_id=external_id).exclude(
        location_name='').values_list('location_name', flat=True).distinct()
    for location in asset_locations:
        asset_summary.append({
            "location_name": location,
            "count":  EveAsset.objects.filter(entity__external_id=external_id, location_name=location).count(),
        })
    return render(
        request,
        'django_eveonline_connector/adminlte/characters/view_character_assets.html',
        context={
            'character': character,
            'asset_summary': asset_summary,
        }
    )


@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
@permission_required('django_eveonline_connector.view_evejumpclone', raise_exception=True)
def view_character_clones(request, external_id):
    character = EveCharacter.objects.get(external_id=external_id)
    return render(
        request,
        'django_eveonline_connector/adminlte/characters/view_character_clones.html',
        context={
            'character': character,
        }
    )


@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
@permission_required('django_eveonline_connector.view_evecontract', raise_exception=True)
def view_character_contracts(request, external_id):
    
    character = EveCharacter.objects.get(external_id=external_id)
    contract_summary = []
    contract_entities = set(list(EveContract.objects.filter(entity__external_id=external_id).values_list(
        'assignee_name')) + list(EveContract.objects.filter(entity__external_id=external_id).values_list('issuer_name')))
    for name in contract_entities:
        name = name[0]
        try:
            related_character = EveCharacter.objects.get(name=name)
            is_alt = related_character.token.user == character.token.user
        except Exception as e:
            is_alt = False 

        valid_contracts = EveContract.objects.filter(
            Q(entity__external_id=external_id) & (Q(assignee_name=name) | Q(issuer_name=name)))
        
        if name == character.name:
            continue

        # resolve ID from database
        contract = EveContract.objects.filter(Q(assignee_name=name)).first()
        if not contract:
            contract = EveContract.objects.filter(
                Q(issuer_name=name)).first()
            summary_entity_id = contract.issuer_id
        else:
            summary_entity_id = contract.assignee_id

        contract_summary.append({
            "name": name, 
            "external_id": summary_entity_id,
            "count": valid_contracts.count(),
            "most_common": valid_contracts.values("type").annotate(count=Count('type')).order_by("-count")[0]['type'].title().replace("_", " "),
            "is_alt": is_alt
        })

    print(contract_entities)
    return render(
        request,
        'django_eveonline_connector/adminlte/characters/view_character_contracts.html',
        context={
            'character': character,
            'contract_summary': contract_summary
        }
    )


@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
@permission_required('django_eveonline_connector.view_evecontact', raise_exception=True)
def view_character_contacts(request, external_id):
    character = EveCharacter.objects.get(external_id=external_id)
    return render(
        request,
        'django_eveonline_connector/adminlte/characters/view_character_contacts.html',
        context={
            'character': character,
        }
    )


@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
@permission_required('django_eveonline_connector.view_eveskill', raise_exception=True)
def view_character_skills(request, external_id):
    character = EveCharacter.objects.get(external_id=external_id)
    character_skills = EveSkill.objects.filter(entity=character)
    return render(
        request,
        'django_eveonline_connector/adminlte/characters/view_character_skills.html',
        context={
            'character': character,
            'skills': character_skills,
            'skill_names': ",".join([skill.skill_name for skill in character_skills]),
            'skill_levels': ",".join([str(skill.trained_skill_level) for skill in character_skills]),
            'groups': set([skill.skill_group for skill in character_skills]),
        }
    )


@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
@permission_required('django_eveonline_connector.view_evejournalentry', raise_exception=True)
def view_character_journal(request, external_id):
    character = EveCharacter.objects.get(external_id=external_id)
    journal_summary = []
    journal_entities = set(list(EveJournalEntry.objects.filter(entity__external_id=external_id, ref_type__in=["player_trading", "player_donation"]).values_list(
        'first_party_name')) + list(EveJournalEntry.objects.filter(entity__external_id=external_id, ref_type__in=["player_trading", "player_donation"]).values_list('second_party_name')))
    for name in journal_entities:
        name = name[0]
        try:
            related_character = EveCharacter.objects.get(name=name)
            is_alt = related_character.token.user == character.token.user
        except Exception as e:
            is_alt = False

        valid_entries = EveJournalEntry.objects.filter(
            Q(entity__external_id=external_id) & (Q(first_party_name=name) | Q(second_party_name=name)))

        if name == character.name:
            continue

        # resolve ID from database
        entry = EveJournalEntry.objects.filter(
            Q(first_party_name=name)).first()
        if not entry:
            entry = EveJournalEntry.objects.filter(
                Q(second_party_name=name)).first()
            summary_entity_id = entry.second_party_id
        else:
            summary_entity_id = entry.first_party_id

        journal_summary.append({
            "name": name,
            "external_id": summary_entity_id,
            "count": valid_entries.count(),
            "most_common": valid_entries.values("ref_type").annotate(count=Count('ref_type')).order_by("-count")[0]['ref_type'].title().replace("_", " "),
            "is_alt": is_alt
        })
    return render(
        request,
        'django_eveonline_connector/adminlte/characters/view_character_journal.html',
        context={
            'character': character,
            "journal_summary": journal_summary,
        }
    )


@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
@permission_required('django_eveonline_connector.view_evetransaction', raise_exception=True)
def view_character_transactions(request, external_id):
    character = EveCharacter.objects.get(external_id=external_id)
    return render(
        request,
        'django_eveonline_connector/adminlte/characters/view_character_transactions.html',
        context={
            'character': character,
        }
    )


@login_required
@permission_required('django_eveonline_connector.view_evecharacter', raise_exception=True)
def view_character_audit(request, external_id):
    character = EveCharacter.objects.get(external_id=external_id)
    return render(
        request,
        'django_eveonline_connector/adminlte/characters/view_character_audit.html',
        context={
            'character': character,
        }
    )



from django.test import TestCase, Client, TransactionTestCase
from django.urls import reverse_lazy, reverse
from unittest.mock import patch
from django_eveonline_connector.models import *
from esipy import EsiClient
from pyswagger.io import Response as PySwaggerResponse
import pyswagger
import esipy
from django.apps import AppConfig
import django_eveonline_connector
from django.core.cache import cache
import uuid
from unittest import skip

# MOCK HELPERS 


class MockObject():
    pass


def mock_eve_client_call_character_public_data(*args, **kwargs):
    obj = MockObject()
    obj.data = {
        'name': 'BearThatCares'
    }
    return obj


def mock_eve_client_call_corporation_public_data(*args, **kwargs):
    obj = MockObject()
    obj.data = {
        'name': 'Doomheim',
        'ticker': 'HELL'
    }
    return obj

# ESI
class TestEveClient(TestCase):
    eve_client = None 
    eve_corporation_missing_ceo = None 
    eve_corporation_has_ceo = None
    eve_character = None 
    eve_ceo = None 
    eve_token = None 
    user = None 

    def setUp(self):
        self.eve_client = EveClient.objects.create(esi_callback_url="TEST", 
            esi_client_id="TEST", 
            esi_secret_key="TEST")

        self.user = User.objects.create_user(username="TEST", password="TEST")

        self.eve_token = EveToken.objects.create(
            access_token="123",
            refresh_token="123",
            user=self.user)

        self.eve_token.scopes.set(EveScope.objects.all())
        self.eve_ceo = EveCharacter.objects.create(
            name="TEST CHARACTER",
            external_id=1234,
            token=self.eve_token)

        self.eve_character = EveCharacter.objects.create(
            name="TEST CHARACTER",
            external_id=123)

        self.eve_corporation_missing_ceo = EveCorporation.objects.create(
            name="TEST",
            external_id=321,
            ticker="TEST")

        self.eve_corporation_has_ceo = EveCorporation.objects.create(
            name="TEST",
            external_id=421,
            ticker="TEST",
            ceo=self.eve_ceo)
    
    def tearDown(self):
        EveClient.objects.all().delete()

    def test_eve_client_save(self):
        self.eve_client.pk = 5 
        self.eve_client.save() 
        self.eve_client = EveClient.objects.all()[0]


    @patch('esipy.EsiClient.request')
    @patch('django_eveonline_connector.models.EveClient.get_esi_client')
    def test_eve_client_call_no_security(self, mock_get_esi_client, mock_esi_request):
        mock_get_esi_client.return_value=EsiClient(headers={'User-Agent': "Krypted Platform"})
        mock_esi_request.return_value=PySwaggerResponse(op='characters_character_id')
        response = self.eve_client.call('characters_character_id', character_id=634915984)
        self.assertIsInstance(response, PySwaggerResponse)

    def test_eve_client_call_required_scopes_character_id(self):
        from django_eveonline_connector.exceptions import EveMissingScopeException
        self.assertRaises(EveMissingScopeException, self.eve_client.call,
                          'characters_character_id_standings', character_id=634915984)
    
    def test_eve_client_call_required_scopes_corporation_id_does_not_exist(self):
        self.assertRaises(Exception, self.eve_client.call, 'characters_character_id_standings', corporation_id=98479678)

    def test_eve_client_call_required_scopes_corporation_id_without_ceo(self):
        from django_eveonline_connector.exceptions import EveMissingScopeException
        self.assertRaises(EveMissingScopeException, self.eve_client.call, 'characters_character_id_standings',
                          corporation_id=self.eve_corporation_missing_ceo.external_id)

    @patch('django_eveonline_connector.models.EveClient.get_esi_client')
    @patch('django_eveonline_connector.models.EveToken.refresh')
    @patch('esipy.EsiClient.request') 
    def test_eve_client_call_required_scopes_corporation_id_with_ceo(self, mock_esi_request, mock_token_refresh, mock_get_esi_client):
        op='corporations_corporation_id_members'
        mock_esi_request.return_value=PySwaggerResponse(op=op)
        mock_token_refresh.return_value=True
        mock_get_esi_client.return_value=EsiClient(headers={'User-Agent': "Krypted Platform"})
        response = self.eve_client.call(op, corporation_id=self.eve_corporation_has_ceo.external_id)
        self.assertIsInstance(response, PySwaggerResponse)

    def test_eve_client_get_instance(self):
        client = EveClient.get_instance() 
        self.assertTrue(client.esi_callback_url == self.eve_client.esi_callback_url)

    @patch('django.apps.apps.get_app_config')
    def test_eve_client_get_instance_app_config(self, mock_app_config):
        mock_app_config.return_value = AppConfig(app_name='django_eveonline_connector', 
            app_module=django_eveonline_connector)
        mock_app_config.return_value.ESI_CLIENT_ID="MOCKED"
        mock_app_config.return_value.ESI_SECRET_KEY="MOCKED"
        mock_app_config.return_value.ESI_CALLBACK_URL="MOCKED"
        mock_app_config.return_value.ESI_BASE_URL="MOCKED"
        client = EveClient.get_instance() 
        self.assertTrue("MOCKED" == client.esi_callback_url)

    def test_eve_client_get_instance_failed(self):
        self.eve_client.delete()
        self.assertRaises(Exception, EveClient.get_instance)
        self.eve_client = EveClient.objects.create(esi_callback_url="TEST", 
            esi_client_id="TEST", 
            esi_secret_key="TEST")

    def test_eve_client_get_esi_app(self):
        esi_app = self.eve_client.get_esi_app()
        self.assertIsInstance(esi_app, pyswagger.core.App)
        self.assertTrue('esi_app' in cache)
        esi_app = self.eve_client.get_esi_app()
        self.assertIsInstance(esi_app, pyswagger.core.App)

    @patch('django_eveonline_connector.models.EveClient.get_esi_security')
    def test_eve_client_get_esi_client(self, mock_get_esi_security):
        mock_get_esi_security.return_value = EsiSecurity(
            redirect_uri=self.eve_client.esi_callback_url,
            client_id=123,
            secret_key="123",
            headers={'User-Agent': "Krypted Platform"}
        )
        self.assertIsInstance(self.eve_client.get_esi_client(), esipy.EsiClient)
        self.assertIsInstance(self.eve_client.get_esi_client(self.eve_token), esipy.EsiClient)

    @patch('esipy.EsiSecurity.update_token')
    def test_eve_client_get_esi_security(self, mock_esi_security_update_token):
        self.assertIsInstance(self.eve_client.get_esi_security(), esipy.EsiSecurity)
        self.assertIsInstance(self.eve_client.get_esi_security(self.eve_token), esipy.EsiSecurity)

# SSO

class TestEveToken(TestCase):
    eve_token = None 
    user = None 

    def setUp(self):
        self.user = User.objects.create_user(
            username="TEST", 
            password="TEST")

        self.eve_token = EveToken.objects.create(
            access_token="TEST",
            refresh_token="TEST",
            user=self.user)

        self.eve_token.scopes.add(EveScope.objects.get_or_create(name="publicData")[0])

        self.eve_character = EveCharacter.objects.create(
            external_id=1,
            name="BearThatTests")
        
    def tearDown(self):
        self.user.delete() 
        self.eve_token.delete() 
        self.eve_character.delete() 

    def test_eve_token_str(self):
        self.assertTrue(self.eve_token.__str__() == "<Unknown Character:%s>" % self.user)
        self.eve_character.token = self.eve_token 
        self.eve_character.save() 
        self.assertTrue(self.eve_token.__str__() == "<%s:%s>" % (self.eve_character.name, self.user))

    @patch('esipy.EsiSecurity.refresh')
    @patch('django_eveonline_connector.models.EveClient.get_esi_security')
    def test_eve_token_refresh(self, mock_get_esi_security, mock_esi_security_refresh):
        mock_get_esi_security.return_value = EsiSecurity(
            redirect_uri="127.0.0.1",
            client_id=123,
            secret_key="123",
            headers={'User-Agent': "Krypted Platform"}
        )
        mock_esi_security_refresh.return_value = {
            'access_token': 'refreshed',
            'refresh_token': 'refreshed',
            'expires_in': 1500,
        }
        self.eve_token.refresh()

        with self.assertLogs('django_eveonline_connector', level='INFO') as cm:
            self.eve_token.refresh()
            self.assertTrue("Token refresh not needed" in cm.output[0])

class TestEveScope(TestCase):
    eve_scope_a = None 
    eve_scope_b = None 
    scopes = None 

    def setUp(self):
        self.eve_scope_a = EveScope.objects.create(name="test_scope_a")
        self.eve_scope_b = EveScope.objects.create(name="test_scope_b")
        self.scopes = EveScope.objects.filter(name__contains="test")

    def tearDown(self):
        self.eve_scope_a.delete() 
        self.eve_scope_b.delete()

    def test_eve_scope_str(self):
        self.assertTrue(self.eve_scope_a.__str__() == self.eve_scope_a.name)

    def test_eve_scope_convert_to_list(self):
        result = EveScope.convert_to_list(self.scopes).sort()
        expected_result = [self.eve_scope_a.name, self.eve_scope_b.name].sort()
        self.assertTrue(result == expected_result)

    def test_eve_scope_convert_to_string(self):
        result = EveScope.convert_to_string(self.scopes)
        expected_result = "%s,%s" % (self.eve_scope_a, self.eve_scope_b)

# ENTITIES
class TestEveEntity(TestCase):
    def test_eve_entity_str(self):
        ee = EveEntity.objects.create(name="TEST", external_id=0)
        self.assertTrue(ee.__str__() == ee.name)
        ee.delete()


class TestEveCharacter(TestCase):
    @patch.object(django_eveonline_connector.models.EveClient, 'call', mock_eve_client_call_character_public_data)
    def test_eve_character_create_from_external_id(self):
        ec = EveCharacter.create_from_external_id(634915984)
        self.assertTrue(ec.name == "BearThatCares")
        self.assertTrue(ec.external_id == 634915984)

class TestEveCorporation(TestCase):
    @patch.object(django_eveonline_connector.models.EveClient, 'call', mock_eve_client_call_corporation_public_data)
    def test_eve_corporation_create_from_external_id(self):
        ec = EveCorporation.create_from_external_id(1000001)
        self.assertTrue(ec.name == "Doomheim")
        self.assertTrue(ec.external_id == 1000001)

class TestEveAlliance(TestCase):
    # TODO: leaving as a sanity check that EveClient.call works.. find better way
    def test_eve_alliance_create_from_external_id(self):
        ea = EveAlliance.create_from_external_id(434243723)
        self.assertTrue(ea.name == "C C P Alliance")
        self.assertTrue(ea.external_id == 434243723)

# ENTITY DATA MODELS
class TestEveAsset(TestCase):
    esi_data_row = {
            "is_blueprint_copy": True,
            "is_singleton": False,
            "item_id": 18,
            "location_flag": "AssetSafety",
            "location_id": 60000004,
            "location_type": "station",
            "quantity": 5,
            "type_id": 1,
        }

    eve_entity = None
    
    def setUp(self):
        self.eve_entity = EveEntity.objects.create(name="TEST", external_id=1)
    
    def tearDown(self):
        self.eve_entity.delete()
        self.eve_entity = None 

    @skip("Fix later")
    @patch('django_eveonline_connector.models.resolve_location_from_location_id_location_type')
    @patch('django_eveonline_connector.models.resolve_type_id_to_category_name')
    @patch('django_eveonline_connector.models.resolve_type_id_to_type_name')
    @patch('django_eveonline_connector.models.resolve_type_id_to_category_id')
    @patch('django_eveonline_connector.models.resolve_type_id_to_group_id')
    def test_eve_asset_create_from_esi_row(self, mock_resolve_type_id_to_group_id, mock_resolve_type_id_to_category_id, mock_resolve_type_id_to_type_name, mock_resolve_type_id_to_category_name, mock_resolve_location_from_location_id_location_type):
        mock_resolve_type_id_to_group_id.return_value = 0
        mock_resolve_type_id_to_category_id.return_value = 0
        mock_resolve_type_id_to_type_name.return_value = "TestTypeName"
        mock_resolve_type_id_to_category_name.return_value = "TestCategoryName"
        mock_resolve_location_from_location_id_location_type.return_value = "Test Station - VII"
        EveAsset.create_from_esi_row(self.esi_data_row, self.eve_entity.external_id)

        eve_asset = EveAsset.objects.all()[0]
        self.assertTrue(eve_asset.is_blueprint_copy == self.esi_data_row['is_blueprint_copy'])
        self.assertTrue(eve_asset.is_singleton == self.esi_data_row['is_singleton'])
        self.assertTrue(eve_asset.item_id == self.esi_data_row['item_id'])
        self.assertTrue(eve_asset.location_flag == self.esi_data_row['location_flag'])
        self.assertTrue(eve_asset.location_id == self.esi_data_row['location_id'])
        self.assertTrue(eve_asset.location_type == self.esi_data_row['location_type'])
        self.assertTrue(eve_asset.type_id == self.esi_data_row['type_id'])
        self.assertTrue(eve_asset.group_id == mock_resolve_type_id_to_group_id.return_value)
        self.assertTrue(eve_asset.category_id == mock_resolve_type_id_to_category_id.return_value)
        self.assertTrue(eve_asset.item_name == mock_resolve_type_id_to_type_name.return_value)
        self.assertTrue(eve_asset.item_type == mock_resolve_type_id_to_category_name.return_value)
        self.assertTrue(eve_asset.location_name == mock_resolve_location_from_location_id_location_type.return_value)
        eve_asset.delete() 
        
        # remove blueprint copy
        self.esi_data_row.pop('is_blueprint_copy')
        EveAsset.create_from_esi_row(self.esi_data_row, self.eve_entity.external_id)
        eve_asset = EveAsset.objects.all()[0]
        self.assertTrue(eve_asset.is_blueprint_copy == False )
        eve_asset.delete() 

        # raise large exception 
        mock_resolve_type_id_to_group_id.side_effect = Exception()
        with self.assertLogs('django_eveonline_connector', level='WARNING') as cm:
            EveAsset.create_from_esi_row(self.esi_data_row, self.eve_entity.external_id)
            self.assertTrue("Failed to create" in cm.output[0])
        self.assertTrue(EveAsset.objects.all().count() == 0)

        
    def test_eve_asset_get_bad_asset_category_ids(self):
        self.assertTrue(EveAsset.get_bad_asset_category_ids() == [42, 43])
        
class TestEveJumpClone(TransactionTestCase):
    eve_data_row = {
        "implants": [1,2,3,4,5],
        "jump_clone_id": 56565496,
        "location_id": 1030491495855,
        "location_type": "structure",
        "name": ""
        }
    
    def setUp(self):
        EveEntity.objects.create(name="TEST_EJC", external_id=2)
    
    def tearDown(self):
        EveEntity.objects.all().delete()

    @patch('django_eveonline_connector.models.resolve_type_id_to_type_name')
    @patch('django_eveonline_connector.models.resolve_location_from_location_id_location_type')
    def test_create_from_esi_row(self, mock_resolve_location_from_location_id_location_type, mock_resolve_type_id_to_type_name):
        eve_entity = EveEntity.objects.get(name="TEST_EJC")
        mock_resolve_location_from_location_id_location_type.return_value = "My Location"
        mock_resolve_type_id_to_type_name.return_value = "Random Implant"
        EveJumpClone.create_from_esi_row(self.eve_data_row, eve_entity.external_id)

        expected_implant_string = "a,a,a,a,a".replace("a", mock_resolve_type_id_to_type_name.return_value)
        jump_clone = EveJumpClone.objects.all()[0]
        self.assertTrue(jump_clone.location_id == self.eve_data_row['location_id'])
        self.assertTrue(jump_clone.location_type == self.eve_data_row['location_type'])
        self.assertTrue(jump_clone.jump_clone_id == self.eve_data_row['jump_clone_id'])
        self.assertTrue(jump_clone.location == mock_resolve_location_from_location_id_location_type.return_value)
        self.assertTrue(jump_clone.implants == expected_implant_string)

# OTHER
class TestEveGroupRule(TestCase):
    def setUp(self):
        self.group = Group.objects.create(name="TEST")
        self.eve_group_rule = EveGroupRule.objects.create(group=self.group)
        self.user = User.objects.create(username="TEST USER")
        self.excluded_user = User.objects.create(username="EXCLUDED USER")
        self.eve_token = EveToken.objects.create(user=self.user) 
        self.eve_alliance = EveAlliance.objects.create(
            name="TEST ALLIANCE PLEASE IGNORE",
            external_id=912345
        )
        self.eve_corporation = EveCorporation.objects.create(
            name="TEST CORPORATION",
            alliance=self.eve_alliance,
            external_id=512345
        )
        self.eve_character = EveCharacter.objects.create(name="TEST", 
            token=self.eve_token, 
            corporation=self.eve_corporation,
            external_id=112345)

    def tearDown(self):
        self.group.delete()
        self.eve_group_rule.delete()
        self.eve_token.delete()
        self.eve_alliance.delete()
        self.eve_corporation.delete()
        self.eve_character.delete()

    def test_eve_group_rule_character_only(self):
        pre_valid_user_list = self.eve_group_rule.valid_user_list
        self.assertEqual(0, pre_valid_user_list.count())
        self.eve_group_rule.characters.add(self.eve_character)
        valid_user_list = self.eve_group_rule.valid_user_list
        self.assertEqual(1, valid_user_list.count())

    def test_eve_group_rule_corporation_only(self):
        self.eve_group_rule.corporations.add(self.eve_corporation)
        valid_user_list = self.eve_group_rule.valid_user_list
        self.assertEqual(1, valid_user_list.count())

    def test_eve_group_rule_alliance_only(self):
        self.eve_group_rule.alliances.add(self.eve_alliance)
        valid_user_list = self.eve_group_rule.valid_user_list
        self.assertEqual(1, valid_user_list.count())

    def test_eve_group_rule_compound(self):
        self.eve_group_rule.characters.add(self.eve_character)
        self.eve_group_rule.corporations.add(self.eve_corporation)
        self.eve_group_rule.alliances.add(self.eve_alliance)
        valid_user_list = self.eve_group_rule.valid_user_list
        self.assertEqual(1, valid_user_list.count())
        self.eve_character.corporation = None 
        self.eve_character.save() 
        valid_user_list = self.eve_group_rule.valid_user_list
        self.assertEqual(0, valid_user_list.count())

    def test_eve_group_rule_compound_with_role(self):
        self.eve_group_rule.characters.add(self.eve_character)
        self.eve_group_rule.roles.add(EveCorporationRole.objects.get(name="Hangar Take 1"))
        self.assertEqual(0, self.eve_group_rule.valid_user_list.count())
        self.eve_character.roles.add(
            EveCorporationRole.objects.get(name="Hangar Take 1"))
        self.assertEqual(1, self.eve_group_rule.valid_user_list.count())

    def test_invalid_user_list(self):
        self.excluded_user.groups.add(self.group)
        self.user.groups.add(self.group)
        self.eve_group_rule.characters.add(self.eve_character)
        self.assertEqual(1, self.eve_group_rule.invalid_user_list.count())


# Generated by Django 2.2.10 on 2020-03-08 20:49

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0023_evecontact_evecontract_evejournalentry_eveskill_evetransaction'),
    ]

    operations = [
        migrations.RenameField(
            model_name='eveasset',
            old_name='location',
            new_name='location_name',
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-27 23:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0020_auto_20200227_2151'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eveasset',
            name='entity',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_eveonline_connector.EveEntity'),
        ),
        migrations.CreateModel(
            name='EveClone',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('location_id', models.IntegerField()),
                ('location_type', models.CharField(max_length=64)),
                ('jump_clone_id', models.IntegerField()),
                ('location', models.CharField(max_length=128)),
                ('implants', models.TextField()),
                ('entity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_eveonline_connector.EveEntity')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.conf import settings
from django.contrib import admin
from django.urls import path, re_path
from django_eveonline_connector.views import sso, character, corporation, api

# SSO 
urlpatterns = [
    path('sso/callback', sso.sso_callback,
         name="django-eveonline-connector-sso-callback"),
    path('sso/token/add/', sso.add_sso_token,
         name="django-eveonline-connector-sso-token-add"),
    path('sso/token/update/<int:token_id>/', sso.update_sso_token,
         name="django-eveonline-connector-sso-token-update"),
    path('sso/token/remove/<int:pk>/', sso.remove_sso_token,
         name="django-eveonline-connector-sso-token-remove"),

]

# Characters
urlpatterns += [
    path('characters/', character.list_characters,
         name="django-eveonline-connector-characters-view"),
    path('characters/select-primary/', character.select_primary_character,
         name="django-eveonline-connector-character-select-primary"),
    path('characters/select-character/', character.select_character,
         name="django-eveonline-connector-character-select"),
    path('character/<int:character_id>/set-primary/', character.set_primary_character,
         name="django-eveonline-connector-character-set-primary"),
    path('character/<int:external_id>/refresh/', character.refresh_character_public,
         name='django-eveonline-connector-character-refresh-public'),
    path('character/view/<int:external_id>/', character.view_character,
         name="django-eveonline-connector-view-character"),
    path('character/refresh/<int:external_id>/', character.refresh_character,
         name="django-eveonline-connector-refresh-character"),
    path('character/view/<int:external_id>/assets/', character.view_character_assets,
         name="django-eveonline-connector-view-character-assets"),
    path('character/view/<int:external_id>/clones/', character.view_character_clones,
         name="django-eveonline-connector-view-character-clones"),
    path('character/view/<int:external_id>/contacts/', character.view_character_contacts,
         name="django-eveonline-connector-view-character-contacts"),
    path('character/view/<int:external_id>/contracts/', character.view_character_contracts,
         name="django-eveonline-connector-view-character-contracts"),
    path('character/view/<int:external_id>/skills/', character.view_character_skills,
         name="django-eveonline-connector-view-character-skills"),
    path('character/view/<int:external_id>/journal/', character.view_character_journal,
         name="django-eveonline-connector-view-character-journal"),
    path('character/view/<int:external_id>/transactions/', character.view_character_transactions,
         name="django-eveonline-connector-view-character-transactions"),
    path('character/view/<int:external_id>/audit/', character.view_character_audit,
         name="django-eveonline-connector-view-character-audit"),
]

# Corporation 
urlpatterns += [
    path('corporations/', corporation.view_corporations,
         name="django-eveonline-connector-corporations-view"),
    path('corporation/view/<int:external_id>/', corporation.view_corporation,
         name="django-eveonline-connector-view-corporation"),
    path('corporation/view/<int:external_id>/structures/', corporation.view_corporation_structures,
         name="django-eveonline-connector-view-corporation-structures"),
    # path('corporation/refresh/<int:external_id>/', character.refresh_corporation,
    #      name="django-eveonline-connector-refresh-corporation"),
]

# Alliance
# urlpatterns += [
#     path('alliance/view/<int:external_id>/', character.view_alliance,
#          name="django-eveonline-connector-view-alliance"),
# ]

# Utilities
urlpatterns += [
    path('api/entity/resolve', api.get_entity_info,
        name="django-eveonline-connector-api-get-entity-info"),
]

# JSON
urlpatterns += [
    path('api/assets/', api.get_assets,
         name="django-eveonline-connector-api-assets"),
    path('api/clones/', api.get_clones,
         name="django-eveonline-connector-api-clones"),
    path('api/contacts/', api.get_contacts,
         name="django-eveonline-connector-api-contacts"),
    path('api/contracts/', api.get_contracts,
         name="django-eveonline-connector-api-contracts"),
    path('api/journal/', api.get_journal,
         name="django-eveonline-connector-api-journal"),
    path('api/transactions/', api.get_transactions,
         name="django-eveonline-connector-api-transactions"),
]



# Generated by Django 2.2.13 on 2020-09-20 22:05

from django.db import migrations

def load_required_migrations(apps, schema_editor):
    EveScope = apps.get_model('django_eveonline_connector', 'EveScope')
    required_scopes = ['publicData', 'esi-contracts.read_character_contracts.v1', 'esi-mail.read_mail.v1', 'esi-skills.read_skills.v1', 'esi-skills.read_skillqueue.v1', 'esi-wallet.read_character_wallet.v1', 'esi-characters.read_contacts.v1',
                       'esi-corporations.read_corporation_membership.v1', 'esi-assets.read_assets.v1', 'esi-clones.read_clones.v1', 'esi-markets.read_character_orders.v1', 'esi-characters.read_corporation_roles.v1', 'esi-universe.read_structures.v1', 'esi-characters.read_standings.v1']
    EveScope.objects.all().delete()
    for scope in required_scopes:
        if EveScope.objects.filter(name=scope).exists():
            EveScope.objects.update(name=scope, required=True)
        else:
            EveScope.objects.create(name=scope, required=True)

def unload_required_migrations(apps, schema_editor):
    EveScope = apps.get_model('django_eveonline_connector', 'EveScope')
    EveScope.objects.all().delete() # dangerous but it'll be fine 

class Migration(migrations.Migration):

    dependencies = [
        ('django_eveonline_connector', '0028_auto_20200920_2202'),
    ]

    operations = [
        migrations.RunPython(load_required_migrations, reverse_code=unload_required_migrations)
    ]
